
// Solar Coaster - depict Sol and Luna rise and set (inspired by the Apple Watch face Solar).
//
// Solar Coaster's algorithms are accurate at all longitudes, and latitudes between -81° south and +82° north.
//
// Many thanks to Vladimir Agafonkin for his GitHub Javascript code SunCalc that provides all the computational machinery.
// During testing I found these minor SunCalc limitations:
//
// - fails in northern hemisphere at latitudes >= +88 with "illegal date" for sun times.
// - fails in southern hemisphere at latitudes >= -87 with "illegal date" for sun times.
// - fails to detect moon always down, "alwaysDown" not set and moonset time not defined.
//
// And thanks to Brian Tingle and his GitHub Javascript code moon-phase. From his 1,000 line project I distilled the
// salient 25 lines of code required to draw Luna in its proper phase. I then added an SVG transform on that object to
// rotate the phase angle in realtime.

const version = "4.0";

// MARK: - Sun constructor

function Sun( name, mass, radius, spectralClass, metallicity, temperature, luminosity, visualMagnitude ) { // Sun constructor
    
    this.name = name;
    this.mass = mass; // kg
    this.radius = radius; // km
    this.spectralClass = spectralClass;
    this.metallicity = metallicity;
    this.temperature = temperature; // K
    this.luminosity = luminosity; // W (J/s)
    this.visualMagnitude = visualMagnitude;
    
    this.previousSpeed = 0; // current - previous > 0 if trending faster (km/dy)
    this.speedTrend = '-';
    this.previousDistance = 0; // current - previous > 0 if trending larger (km)
    this.distanceTrend = '-';

    this.angularDiameterFromEarth = function ( now ) {
        var [d, trend] = this.earthDistance( now );
        var ad = 2 * Math.asin( ( 2 * this.radius ) / ( 2 * d ) );
        console.log('d',d,'ad',ad,'r',this.radius);
        return ( ad * 180 ) / Math.PI;
   } // end angularDiameterFromEarth

    const earthSunA = 149597870.7; // semi-major axis km, 1 AU
    const earthSunP = 365.25636; //  period dy, 1 sidereal yr
    const earthSunE = 0.0167086; // eccentricity
    const earthSunB = 449; // barycenter km
    
    this.earthAngularSpeed = function ( r ) { // Sun's angular speed relative to Earth at distance r km, in degrees/dy
        var omega = ( (360 / earthSunP) * (earthSunA / r) ) * Math.sqrt( (2 * (earthSunA / r)) - 1 );
        return omega;
    } // end earthAngularSpeed
    
    this.earthOrbitalSpeed = function ( r ) { // Sun's linear speed relative to Earth at distance r km, in km/dy
        var s = ((2 * Math.PI * earthSunA) / earthSunP) * Math.sqrt( (2 * (earthSunA / r) - 1) );
        var deltaSpeed = s - this.previousSpeed;
        if ( deltaSpeed != 0 ) {
            this.speedTrend = ( deltaSpeed == 0 ? "-" : ( deltaSpeed > 0 ) ? unicodeUpTriangle : unicodeDownTriangle );
            this.previousSpeed = s;
        }
        return [s, this.speedTrend];
    } // end earthOrbitalSpeed
    
    this.earthDistance = function ( now ) { // Earth - Sun distance r, in km
        // r = a * ( 1 - e*e ) / ( 1 + e * cos(θ) )
        var fullYear = now.getFullYear()
        var perihelion = new Date( fullYear, 0, 3, 8, 0, 0, 0 );          // average 3 Jan 0800, the default
        var periList = {                                                  // from the US Naval Observatory
            '2017'  :     'Jan   4 14 18',
            '2018'  :     'Jan   3 05 35',
            '2019'  :     'Jan   3 05 20',
            '2020'  :     'Jan   5 07 48',
            '2021'  :     'Jan   2 13 51',
            '2022'  :     'Jan   4 06 55',
            '2023'  :     'Jan   4 16 17',
            '2024'  :     'Jan   3 00 39',
            '2025'  :     'Jan   4 13 28',
            '2026'  :     'Jan   3 17 16',
            '2027'  :     'Jan   3 02 33',
            '2028'  :     'Jan   5 12 28',
            '2029'  :     'Jan   2 18 13',
            '2030'  :     'Jan   3 10 12',
            '2031'  :     'Jan   4 20 48',
            '2032'  :     'Jan   3 05 11',
            '2033'  :     'Jan   4 11 51',
            '2034'  :     'Jan   4 04 47',
            '2035'  :     'Jan   3 00 54',
            '2036'  :     'Jan   5 14 17',
            '2037'  :     'Jan   3 04 00',
            '2038'  :     'Jan   3 05 01',
            '2039'  :     'Jan   5 06 41',
            '2040'  :     'Jan   3 11 33',
            '2041'  :     'Jan   3 21 52',
            '2042'  :     'Jan   4 09 07',
            '2043'  :     'Jan   2 22 15',
            '2044'  :     'Jan   5 12 52',
            '2045'  :     'Jan   3 14 56',
            '2046'  :     'Jan   3 00 58',
            '2047'  :     'Jan   5 11 44',
            '2048'  :     'Jan   3 18 05',
            '2049'  :     'Jan   3 10 27',
            '2050'  :     'Jan   4 19 35',
            '2051'  :     'Jan   3 05 32',
            '2052'  :     'Jan   5 09 18',
            '2053'  :     'Jan   3 22 18',
            '2054'  :     'Jan   2 17 59',
            '2055'  :     'Jan   5 12 24',
            '2056'  :     'Jan   4 03 44',
            '2057'  :     'Jan   3 03 11',
            '2058'  :     'Jan   5 04 00',
            '2059'  :     'Jan   3 10 44',
            '2060'  :     'Jan   4 22 56',
            '2061'  :     'Jan   4 07 33',
            '2062'  :     'Jan   2 21 16',
            '2063'  :     'Jan   5 13 41',
            '2064'  :     'Jan   4 11 54',
            '2065'  :     'Jan   2 15 10',
            '2066'  :     'Jan   5 05 45',
            '2067'  :     'Jan   3 17 05',
            '2068'  :     'Jan   4 14 29',
            '2069'  :     'Jan   4 18 25',
            '2070'  :     'Jan   3 02 51',
            '2071'  :     'Jan   5 09 44',
            '2072'  :     'Jan   4 21 02',
            '2073'  :     'Jan   2 14 41',
            '2074'  :     'Jan   5 11 25',
            '2075'  :     'Jan   4 02 14',
            '2076'  :     'Jan   3 23 33',
            '2077'  :     'Jan   4 20 53',
            '2078'  :     'Jan   3 04 23',
            '2079'  :     'Jan   5 00 15',
            '2080'  :     'Jan   5 07 55',
            '2081'  :     'Jan   2 18 56',
            '2082'  :     'Jan   5 13 51',
            '2083'  :     'Jan   4 11 28',
            '2084'  :     'Jan   3 16 49',
            '2085'  :     'Jan   5 04 33',
            '2086'  :     'Jan   3 15 22',
            '2087'  :     'Jan   4 16 38',
            '2088'  :     'Jan   5 14 59',
            '2089'  :     'Jan   2 17 01',
            '2090'  :     'Jan   5 04 45',
            '2091'  :     'Jan   4 19 22',
            '2092'  :     'Jan   3 15 29',
            '2093'  :     'Jan   5 10 16',
            '2094'  :     'Jan   4 00 17',
            '2095'  :     'Jan   4 03 56',
            '2096'  :     'Jan   5 21 56',
            '2097'  :     'Jan   3 00 52',
            '2098'  :     'Jan   5 02 34',
            '2099'  :     'Jan   5 06 18',
            '2100'  :     'Jan   3 13 54',
        };
        for ( var key in periList ) {                                     // refine perihelion if possible
            if ( periList.hasOwnProperty( key ) ) {
                if ( key == fullYear ) {
                    var toks = periList[ key ].split( / +/ );
                    perihelion = new Date( fullYear, 0, toks[1], toks[2], toks[3], 0, 0, 0 );
                    break;
                }
            }
        }
        var days = ( now - perihelion ) / 1000.0 / 86400.0;               // milliseconds -> seconds -> days since perihelion
        var degrees = days / ( earthSunP / 360.0 );                           // degrees along the orbital ellipse
        var radians = ( degrees * Math.PI ) / 180.0;                      // radians for the Math routines
        var unitsFactor = ( 1 ? 1.0 : 1.609344 );                         // TRUE for km, FALSE for miles
        //
        //var solarDist = ( ( a * ( 1 - e*e ) ) / ( 1 + e * Math.cos( radians ) ) - earthSunB ) / unitsFactor;    // distance
        //
        // I've had to split the above statement:  Xcode parser barfs, apparantly because the unitsFactor divide symbol is confused with a regexp.
        // The result is that the JS pulldown function list is truncated and it's not possible to navigate this source code.
        var earthDist = ( ( earthSunA * ( 1 - earthSunE * earthSunE ) ) / ( 1 + earthSunE * Math.cos( radians ) ) - earthSunB );
        earthDist /= unitsFactor;    // distance
        var deltaDistance = earthDist - this.previousDistance;
        if ( deltaDistance != 0 ) {
            this.distanceTrend = ( deltaDistance == 0 ? "-" : ( deltaDistance > 0 ) ? unicodeUpTriangle : unicodeDownTriangle );
            this.previousDistance = earthDist;
        }
        return [earthDist, this.distanceTrend];
    } // end earthDistance
    
} // end Sun

// MARK: - Moon constructor

function Moon( name, mass, radius ) { // Moon constructor
    
    this.name = name;
    this.mass = mass; // kg
    this.radius = radius; // km
    
    this.previousSpeed = 0; // current - previous > 0 if trending faster (km/dy)
    this.speedTrend = '-';
    this.previousDistance = 0; // current - previous > 0 if trending larger (km)
    this.distanceTrend = '-';

    this.angularDiameterFromEarth = function ( now ) {
        var [d, trend] = this.earthDistance( now );
        var ad = 2 * Math.asin( ( 2 * this.radius ) / ( 2 * d ) );
        return ( ad * 180 ) / Math.PI;
   } // end angularDiameterFromEarth

    const earthMoonA = 384399; // semi-major axis km
    const earthMoonP = 27.321661; // period dy, 1 sidereal month
    const earthMoonE = 0.0549; // eccentricity
    const earthMoonB = 1700; // barycenter km

    this.earthAngularSpeed = function ( r ) { // Moon's angular speed relative to Earth at distance r km, in degrees/dy
        var omega = ( (360 / earthMoonP) * (earthMoonA / r) ) * Math.sqrt( (2 * (earthMoonA / r)) - 1 );
        return omega;
    } // end earthAngularSpeed
    
    this.earthOrbitalSpeed = function ( r ) { // Moon's linear speed relative to Earth at distance r km, in km/dy
        var s = ((2 * Math.PI * earthMoonA) / earthMoonP) * Math.sqrt( (2 * (earthMoonA / r) - 1) );
        var deltaSpeed = s - this.previousSpeed;
        if ( deltaSpeed != 0 ) {
            this.speedTrend = ( deltaSpeed == 0 ? "-" : ( deltaSpeed > 0 ) ? unicodeUpTriangle : unicodeDownTriangle );
            this.previousSpeed = s;
        }
        return [s, this.speedTrend];
    } // end earthOrbitalSpeed
    
    this.earthDistance = function ( now ) { // Earth - Moon distance r, in km
        var moonPos = SunCalc.getMoonPosition( now, latitude, longitude);
        var deltaDistance = moonPos.distance - this.previousDistance;
        if ( deltaDistance != 0 ) {
            this.distanceTrend = ( deltaDistance == 0 ? "-" : ( deltaDistance > 0 ) ? unicodeUpTriangle : unicodeDownTriangle );
            this.previousDistance = moonPos.distance;
        }
        return [moonPos.distance, this.distanceTrend];
    } // end earthDistance
    
    this.phase = function ( now ) { // current moon phase
        var lunaIllumination = SunCalc.getMoonIllumination( now );
        var phase = lunaIllumination.phase;
        phase = parseFloat( phase ).toFixed( 2 );
        var phaseHuman = "UnknownPhase";
        if ( phase == '0.00' ) {
            phaseHuman = 'New';
        } else if ( phase == '0.25' ) {
            phaseHuman = 'First Quarter';
        } else if ( phase == '0.50' ) {
            phaseHuman = 'Full';
        } else if ( phase == '0.75' ) {
            phaseHuman = 'Last Quarter';
        } else {
            phase = lunaIllumination.phase;
            if ( phase > 0.00 && phase < 0.25 ) {
                phaseHuman = 'Waxing Crescent';
            } else if ( phase > 0.25 && phase < 0.50 ) {
                phaseHuman = 'Waxing Gibbous';
            } else if ( phase > 0.50 && phase < 0.75 ) {
                phaseHuman = 'Waning Gibbous';
            } else if ( phase > 0.75 ) {
                phaseHuman = 'Waning Crescent';
            } else {
                phaseHuman = 'unknown phase';
            }
        }
        var illumination = parseFloat( lunaIllumination.fraction * 100.0 ); // %
        return phaseHuman + " " + illumination.toFixed( illumination >= 100 ? 0 : 1 ) + '%';
    } // end phase
    
} // end Moon

// MARK: - Globals

const Sol                 = new Sun( 'Sol', 1.9885e+30, 6.95700e+5,  'G2 V', 0.0122, 5778, 3.828e+26, -26.74);
const Luna                = new Moon( 'Luna', 7.347e+22, 1.7374e+3);
const unicodeUpTriangle   = '\u21e1';
const unicodeDownTriangle = '\u21e3';
const eType               = {
    'P' : 'Partial',
    'T' : 'Total',
    'N' : 'Penumbral',
};
const eTypeSolar          = {
    'P' : 'Partial',
    'A' : 'Annular',
    'T' : 'Total',
};
const fullMonth1          = {
    'Jan' : 'January',
    'Feb' : 'February',
    'Mar' : 'March',
    'Apr' : 'April',
    'May' : 'May',
    'Jun' : 'June',
    'Jul' : 'July',
    'Aug' : 'August',
    'Sep' : 'September',
    'Oct' : 'October',
    'Nov' : 'November',
    'Dec' : 'December',
};
const youDriveNone           = 0;
const youDriveSpecialMinutes = 1;
const youDriveAllMinutes     = 2;

var todayExt;               // 1 IFF running as an iOS Today extension
var screenSaver;            // 1 IFF running as a macOS ScreenSaver
var minutesInterval;        // minutes between updates
var sinePath;               // Sol's SVG tracking path
var sinePathLen;            // length of path
var horizon;                // SVG line
var sol;                    // the Sun
var solRadius;              // in pixels
var solHalo;                // Gaussian blurred cicle that tracks Sol
var sunriseTime;            // human form
var sunsetTime;             // human form
var sunrise;                // minute of day
var sunset;                 // minute of day
var dawn;                   // minute of day
var dusk;                   // minute of day
var noon;                   // minute of day
var solarNoon;              // minute of day
var solarMidnight;          // minute of day
var solarSkew;              // basically the differnece between solar noon and real noon (allows for a horizontal horizon line)
var twiLen;                 // twilight length in minutes
var nightSolColor;          // black or purple
var longitude;              // our longitude
var latitude;               // our latitude
var altitude;               // our height above sea level, meters
var lunaSinePath;           // Luna's SVG tracking path
var lunaSinePathLen;        // length of path
var lunaPath;               // Luna's foreground (arc) path
var lunabackPath;           // Luna's background path
var crescentRotationAngle;  // crescent rotation angle, phase dependant
var moonrise;               // minute of the day
var moonset;                // minute of the day
var youDrive;               // 0 = production, > 0 youDrive: 1 just specialMinutes, > 1 all 1440 minutes
var debug;                  // general debugging flag
var debugDate;              // empty string, else a particular date to simulate, such as '2017-03-27'
var debugDateOrig;          // after an eclipse reset resume with this date rather than ""
var specialMinutes;         // sunrise, sunset, solar noon, whatever, for debugging
var showDetails;            // true to show the gore
var title;                  // show details title element
var subtitle;               // show details subtitle element
var titleTime;              // show details time element
var mpd;                    // minutes per day
var timerHandle;            // so we can cancel timer
var newDayInProgress;       // do not show Luna during this transition
var lastKnownMinute;        // replay lost minutes counter
var use24HourClock;         // 12 or 24 hour clock format
var tweenTime = -1;         // -1 when not tweening, 0 otherwise (the tween start time)
var tweenStart;             // start minute of the simulated day
var tweenEnd;               // end minute of the simulated day
var tweenTimeEnd;           // either 1 or 2 seconds (the tween end time)
var tweenDT;                // milliseconds between tweens
var trendingDayLength;      // > 0 if trending longer (seconds)
var previousDayLength;      // current - previous > 0 if trending longer (seconds)
var currentMinute;          // minute currently being displayed
var eclipses;               // list of eclipses for this year and next
var eclipsesVisible;        // list of visible eclipses for this year and next
var imminentEclipse;        // circumstances of soonest eclipse, solar or lunar
var lunarEclipses;          // list of visible lunar eclipses this year and next
var myAlertActive;          // TRUE IFF myAlert() has a modal active
var iOSInfoMsg;             // full HTML help info injected here
var backgroundColor;        // current theme color
var darkMode;               // "" for light, "dark" for dark, *** set by Objective-C wrapper ***
var originalModalBackground;// displayInfo() changes this to black
var embeddedFromOutsideAPP; // special configurartion if Outside has embedded us
var sliderObj;              // SVG slider

// MARK: - Initialization

Date.prototype.dst = function() {
    
    return this.getTimezoneOffset() < this.stdTimezoneOffset();
    
} // end Date.dst

Date.prototype.stdTimezoneOffset = function() {
    
    var jan = new Date(this.getFullYear(), 0, 1);
    var jul = new Date(this.getFullYear(), 6, 1);
    return Math.max(jan.getTimezoneOffset(), jul.getTimezoneOffset());
    
} // end Date.stdTimezoneOffset

document.addEventListener("visibilitychange", function() {
    
    if (document.visibilityState === 'visible') {
        resetAnimations();
    } else {
        dismissModal();
    }
    
});

function init1() {
    
    // We factor out this portion of App initialization for use by the newDay() animation.
    
    sinePath = document.getElementById( 'solarSinePath' );
    horizon = document.getElementById( 'solarHorizon' );
    sol = document.getElementById( 'solarSol' );
    solHalo = document.getElementById( 'solarSolHalo' );
    sunriseTime = document.getElementById( 'solarSunriseTime' );
    sunsetTime = document.getElementById( 'solarSunsetTime' );
    moonriseTime = document.getElementById( 'solarMoonriseTime' );
    moonsetTime = document.getElementById( 'solarMoonsetTime' );
    lunaSinePath = document.getElementById('lunaSinePath');
    lunaPath = document.getElementById( 'luna');
    lunabackPath = document.getElementById( 'lunaBack');
    title = document.getElementById( 'title');
    subtitle = document.getElementById( 'subtitle');
    titleTime = document.getElementById( 'titleTime');
    document.getElementById( 'version' ).textContent = 'Solar Coaster ';// +  version;
    sinePathLen = sinePath.getTotalLength();
    lunaSinePathLen = lunaSinePath.getTotalLength();
    solRadius = 45;
    sol.setAttribute( 'r', solRadius );
    solHalo.setAttribute( 'r', ( solRadius * 2 ) * 1.111111 );
    
} // end init1

function init2 () {
    
    // Debug debugDates :)
    
    //debugDate = '2017-08-19'; // e.g. just before solar eclipse day, '2017-08-21'
    //debugDate = '2022-05-15'; // e.g. just before lunar eclipse day
    //debugDate = '2017-08-21'; // e.g. solar eclipse day, '2017-08-21'
    //debugDate = '2044-08-22';
    //debugDate = '2048-06-11';
    //debugDate = '2018-01-31'; // full lunar eclipse
    //debugDate = '2017-11-05'; // e.g. end of dalight savings
    //debugDate = '2018-03-11'; // e.g. start of dalight savings  Ork Solar Midnight error
    //debugDate = '2020-07-05'; // e.g. random lunar eclipse day
    //debugDate = '2017-02-09';
    //debugDate = '2019-01-20';
    //debugDate = '2022-11-08'; // full lunar eclipse SM on the left
    //debugDate = '2023-06-23'; // summer soltice
    //debugDate = '2023-06-15'; // summer soltice
    //debugDate = '2023-01-06'; // capital insurrection + 2 yr

    // Production debugDate.
    
    //debugDate = '';             // Today
    
    debugDateOrig = debugDate;
    
    if ( debugDate ) {
        alert('Running with a fake day '+debugDate+'!');
    }

    debug = 0;
    youDrive = youDriveNone;
    minutesInterval = 1;
    mpd = 24 * 60;
    nightSolColor = 'purple';
    specialMinutes = [];
    showDetails = 0;
    newDayInProgress = 0;
    lastKnownMinute = getNowMinute() - 1;
    use24HourClock = 1;
    tweenTime = -1;
    tweenDT = 10;
    todayExt = 0;
    screenSaver = 0;
    eclipses = [];
    eclipsesVisible = [];
    imminentEclipse = [];
    lunarEclipses = [];
    myAlertActive = 0;
    iOSInfoMsg = '';
    darkMode = '';
    originalModalBackground = '';
    embeddedFromOutsideAPP = 0;

    if ( document.getElementById( 'solarcoaster' ).style['background'] == 'linear-gradient(rgb(3, 166, 243), black)' ) { // running embedded from Outside
        var element = document.getElementById( 'version' );
        element.setAttribute( 'fill', 'blue' );
    }
    setDetailedInfoVisibility( 'hidden' );
    
    var element = document.getElementById( 'solarcoaster' );
    
    // Latitude, longitude and altitude is set variously:
    //
    // 1) hardcoded, if via a simple web invocation.
    // 2) if they are specified as a query string as below, likely only from being embedded by Outside.
    // 3) finally if loaded by Objective-C wrapper code (loadHTMLString), and injected via a call to setLatLonAlt().
    
    var params = {};
    if (location.search) {
        var parts = location.search.substring(1).split('&');
        for (var i = 0; i < parts.length; i++) {
            var nv = parts[i].split('=');
            if (!nv[0]) continue;
            params[nv[0]] = nv[1] || true;
        }
    }
    if ( params.latitude && params.longitude && params.altitude ) {
        latitude =  params.latitude;
        longitude = params.longitude;
        altitude = params.altitude;
        if ( params.Outside ) {
            embeddedFromOutsideAPP = 1;
            setBackground( '#03a6f3' );
            document.getElementById( 'version' ).textContent = 'Solar Coaster ' +  version;
            document.getElementById( 'help').setAttribute( 'visibility', 'hidden' );
        }
        setLatLonAlt(latitude, longitude, altitude);
    } else {
        latitude  =  41.2; // degress north latitude
        longitude = -96.2; // degress west longitude
        altitude = 366; // meters
        element.style['width'] = '560px';
        element.style['height'] = '960px';
        iOSInfoMsg = "<h1><font color=yellow>Psuedo-help for Web Testing</font><h1>";
        //setLatLonAlt( latitude, longitude, altitude );
    }

    initTrendingData();     // set initial "previous" values so we can determine trends later
    updateRiseAndSetTimes ( getNowMinute() ); // update Sol and Luna rise and set times, minute of the day
    makeSlider();
    
    console.log("### Solar Coaster version " + version + " JavaScript / SVG, " + getNowDateObject() + ", lat " + latitude + ", lon " + longitude + " alt " + altitude );
    
} // end init2

function initTrendingData () {
    
    previousDayLength = pdl();
    var nowMinus5Mins = getNowDateObject();
    nowMinus5Mins.setMinutes( nowMinus5Mins.getMinutes() - 10 ); // well, 10 minutes rather than 5 :)
    
    var distance, trend;
    [distance, trend] = Luna.earthDistance( nowMinus5Mins );
    Luna.earthOrbitalSpeed( distance );
    [distance, trend] = Sol.earthDistance( nowMinus5Mins );
    Sol.earthOrbitalSpeed( distance );

} // initTrendingData

window.onload = function () {

    init1();                // global DOM elements
    init2();                // global variables
    resetAnimations();       // required for iOS, seems not to honor onfocus and onblur :(
    
} // end onload

// MARK: - Mainloop

function adjustMarkers ( markers ) {
    
    var newCoord, elem, kind, trans;
    for ( var key in markers ) {
        if ( markers.hasOwnProperty( key ) ) {
            newCoord = getPointForMinute( markers[ key ] - solarSkew );
            elem = document.getElementById( key );
            kind = elem.constructor.name;
            if ( kind == 'SVGCircleElement' ) {
                trans = " translate(" + newCoord.x + "," + newCoord.y + ")";
            } else if ( kind == 'SVGRectElement' ) {
                var newx = newCoord.x - 14;
                var newy = newCoord.y - 14;
                trans = " translate(" + newx + "," + newy + ")";
            } else if ( kind == 'SVGPolygonElement' ) {
                var newx = newCoord.x - 14 -8;
                var newy = newCoord.y - 14 -8;
                trans = " translate(" + newx + "," + newy + ") scale( 0.35 )";
            }
            elem.setAttribute( "transform", trans );
        }
    }
    
} // end adjustMarkers

function animateBodies ( minuteOfDay ) {
    
    var interval = ( youDrive == youDriveNone ? 1000 * 60 * minutesInterval : ( youDrive == youDriveSpecialMinutes ? 2000 : 10 ) );
    
    if ( minuteOfDay == solarMidnight && youDrive != youDriveSpecialMinutes ) {
        if ( ! newDayInProgress ) {
            newDay( solarMidnight  + 1 );
        }
        return;
    }

    if ( youDrive ) {
        var minute, lastMin, nextMin;
        if ( youDrive == youDriveSpecialMinutes && specialMinutes.length >= 1 ) {
            minute = specialMinutes[ minuteOfDay ];
            nextMin = minuteOfDay + 1;
            lastMin = specialMinutes.length;
        } else {
            minute = minuteOfDay;
            nextMin = minute + 1;
            lastMin = mpd;	// day minutes range from 0 -> 1439
            //if ( ( minute >= dawn && minute <= sunrise ) || ( minute >= sunset && minute <= dusk ) ) {
                //interval *= 10;
            //}
        }
        if ( nextMin > lastMin ) {
            animateBodies( 0 ); // new day
            return;
        }

        updateRiseAndSetTimes ( minute ); // update Sol and Luna rise and set times, minute of the day
        animateSol ( minute );
        animateLuna ( minute  );
        moveSlider ( minute );
        timerHandle = setTimeout(  function() { animateBodies( nextMin ) },  interval );
        
    } else {			// not youDrive

        var minute = getNowMinute();
        updateRiseAndSetTimes ( minute ); // update Sol and Luna rise and set times, minute of the day
        if ( minute == ( lastKnownMinute + minutesInterval ) || minute == lastKnownMinute  ) {
            tweenTime = -1;
            lastKnownMinute = minute;
        } else {		// if not now
            minute = playMissedMinutes( minute );
            interval = tweenDT;
        } // ifend tween or not
        animateSol ( minute );
        animateLuna ( minute );
        timerHandle = setTimeout(  function() { animateBodies( solarMidnight + 1 ) }, interval );
        
    } // ifend youDrive
    
} // end animateBodies

function animateDetails () {
    
    var minute = getNowMinute();
    updateRiseAndSetTimes( minute );
    animateSol ( minute );
    animateLuna ( minute );
    timerHandle = setTimeout( function() { animateDetails() }, 1000 * 60 * minutesInterval );
    
} // end animateDetails

function animateLuna ( minute ) {
    
    // Update Luna's phase and visibility.
    
    createPhasedMoon ( minute  ); // draw Luna with phase
    
    var lunaTimeInterval = moonset - moonrise; // in minutes
    var flag = lunaTimeInterval;	       // > 0 if moonset > moonrise, < 0 if moonset < moonrise
    if ( lunaTimeInterval <= 0 ) {
        lunaTimeInterval = mpd + lunaTimeInterval;
    }
    
    var lunaVisible;
    if ( flag > 0 ) {
        lunaVisible = (   ( minute >= moonrise ) && ( minute < moonset ) ) ? 1 : 0;
    } else {
        lunaVisible = (    minute >= moonrise  ||  minute < moonset  ) ? 1 : 0;
    }
    if ( moonrise == -1 && ( minute > moonset || minute < solarSkew ) ) {
        lunaVisible = 0;
    }
    if ( moonset == -1 && ( minute > moonrise || minute < solarSkew ) ) {
        lunaVisible = 1;
    }
    if ( newDayInProgress ) {
        lunaVisible = 0;
    }
    
    var opacity = lunaVisible ? 1.0 : 0.0;
    lunaSinePath.setAttribute( 'opacity', opacity );
    moonriseTime.setAttribute( 'opacity', opacity );
    moonsetTime.setAttribute( 'opacity', opacity );
    // Times are always hidden in v4.0 and greater.
    var vis = 'hidden';
    moonriseTime.setAttribute( 'visibility', vis );
    moonsetTime.setAttribute( 'visibility', vis );

    if ( lunaVisible ) {
        var mins = minute - moonrise;
        if ( flag < 0 && mins < 0 ) {
            mins = mpd - moonrise + minute;
        }
        var newCoord = lunaSinePath.getPointAtLength ( lunaSinePathLen *    ( mins    / lunaTimeInterval )  );
        var trans = " translate(" + ( newCoord.x ) + "," + ( newCoord.y -37.5 ) + ")";
        var bbox = lunabackPath.getBBox();
        var a = crescentRotationAngle;		// crescent rotation angle
        var x = bbox.x + bbox.width / 2;
        var y = bbox.y + bbox.height / 2;
        var rotate = "  rotate(" +  a + "," + x + "," + y +")";
        lunaPath.setAttribute( "transform", trans + ' ' + rotate );
        lunabackPath.setAttribute( "transform", trans + ' ' + rotate );
        setLunaOpacity( lunaTimeInterval, mins ); // vary opacity around moonrise and moonset times
    } else {
        lunaPath.setAttribute( 'opacity', opacity );
        lunabackPath.setAttribute( 'opacity', opacity );
    }
    
    // Adjust various markers.

    adjustMarkers( { "moonriseCircle" : moonrise, "moonsetCircle" : moonset } );
    
} // end animateLuna

function animateSol ( minute ) {
    
    // Move the artificial horizon line to its proper "latitude" and adjust its length.
    
    var horizonCoord = getPointForMinute ( sunrise - solarSkew );
    horizon.setAttributeNS( null, "y1", horizonCoord.y );
    horizon.setAttributeNS( null, "y2", horizonCoord.y );
    var horizonWide = document.getElementById( 'solarHorizonWide' );
    horizonWide.setAttributeNS( null, "y1", horizonCoord.y );
    horizonWide.setAttributeNS( null, "y2", horizonCoord.y );
    var x1 = horizonCoord.x;
    horizonCoord = getPointForMinute( sunset - solarSkew );
    horizon.setAttribute( 'x1', x1 );
    horizon.setAttribute( 'x2', horizonCoord.x );
    
    // Translate the daylight more/less marker.
    
    var triangleName = ( trendingDayLength >= 0 ? 'triangleDn' : 'triangleUp' );
    var triangle = document.getElementById( triangleName );
    triangle.setAttribute( 'visibility', 'visible' );
    if ( triangleName == 'triangleDn' ) {
        document.getElementById( 'triangleUp' ).setAttribute( 'visibility', 'hidden' );
    } else {
        document.getElementById( 'triangleDn' ).setAttribute( 'visibility', 'hidden' );
    }
    var tx = 380;
    var ty = horizonCoord.y - 20;
    triangle.setAttribute( 'transform', 'translate(' + tx + ',' + ty + ')' );
    
    // Adjust You Drive details delta time so that horizon does not overlay it.
    
    if ( horizonCoord.y <= 560 ) {
        titleTime.setAttribute( 'y', 620 )
    } else {
        titleTime.setAttribute( 'y', 550 );
    }
    
    // Compute the path point corresponding to the current time of day and move Sol there, as well as
    // the halo.  Add filters and gradients as required, update title/time.
    
    var newCoord = getPointForMinute( minute - solarSkew );
    var trans = " translate(" + newCoord.x + "," + newCoord.y + ")";
    sol.setAttribute( "transform", trans );
    solHalo.setAttribute( "transform", trans );
    gradientFillSol( minute );
    updateTitle( minute );
    
    // Adjust various markers.
    
    var markers = {
        "dawnCircle"          : dawn,
        "duskCircle"          : dusk,
        "solarNoonCircle"     : solarNoon,
        "noonCircle"          : ( mpd / 2 ),
        "midnightCircle"      : 0,
        "solarMidnightCircle" : solarMidnight,
        "solarMidnightCircleLeft" : solarMidnight + 1,
        "solarMidnightCircleRight" : solarMidnight - 1,
        "sunriseCircle"       : sunrise,
        "sunsetCircle"        : sunset,
        "nowCircle"           : getNowMinute(),
    };
    adjustMarkers( markers );
    
} // end animateSol

function getPointForMinute ( minute ) {
    
    if ( isNaN( minute ) ) {
        minute = 0;
    } else if ( minute < 0 ) {
        minute = minute + mpd;
    } else if ( minute > mpd ) {
        minute = minute - mpd;
    }
    if ( minute > mpd ) {
        console.log("minute="+minute+" > " + mpd);
        minute = 0;
        tweenTime = -1;
    }
    var lengthAlongPath = sinePathLen * ( minute / mpd );
    if ( ! isFinite( lengthAlongPath ) || isNaN( lengthAlongPath )) {
        console.log( "getPointForMinute: lengthAlongPath is not finite or not a number" );
        lengthAlongPath = 0;
    }
    
    if ( youDrive ) {
        if ( lengthAlongPath   > sinePathLen ) {
            lengthAlongPath = sinePathLen;
            console.log("length > pathLen: minute="+minute+", len="+lengthAlongPath+", sinePath="+sinePathLen);
        }
        if ( lengthAlongPath   < 0 ) {
            lengthAlongPath = 0;
            console.log("length < 0: minute="+minute+", len="+lengthAlongPath+", sinePath="+sinePathLen);
        }
    }
    
    return sinePath.getPointAtLength ( lengthAlongPath );
    
} // end getPointForMinute

resetAnimations = function () {
    
    tweenTime = -1; // in case tween never completed
    //console.log("solar Coaster onVisible "+new Date() + ", LNM=" + lastKnownMinute + ", minute=" + getNowMinute() + ", tweenTime="+tweenTime);
    clearTimeout( timerHandle );
    animateBodies( solarMidnight + 1 );
    
} // end resetAnimations

function updateRiseAndSetTimes( minute ) {

    if ( minute == null ) {
        alert( "updateRiseAndSetTimes: minute undefined" );
    }
    currentMinute = minute;
    var hhh = Math.floor (minute / 60.0); // hour of day
    var mmm = minute - (hhh*60);          // minute of day
    var now = getNowDateObject();
    now = new Date( now.getFullYear(), now.getMonth(), now.getDate(), hhh, mmm, 0, 0 );

    var nowDate = document.getElementById( 'solarDate' );
    nowDate.innerHTML = now.getFullYear() + '-' + addZero( now.getMonth() + 1 ) + '-' + addZero( now.getDate() ) +
        ' (' + latitude + ',' + longitude + ',' + altitude + ')'
    
    // Set minute of the day sunrise and sunset times, as well as human format display strings.
    
    var times = SunCalc.getTimes( now, latitude, longitude, altitude );

    sunriseTime.textContent = addZero( times.sunrise.getHours() ) + ':' + addZero( times.sunrise.getMinutes() );
    sunsetTime.textContent = addZero( times.sunset.getHours() ) + ':' + addZero( times.sunset.getMinutes() );
    
    var h = sunriseTime.innerHTML.substring( 0, 2 );
    var m = sunriseTime.innerHTML.substring( 3, 5 );
    sunrise = ( h * 60 ) +  ( m * 1 );
    h = sunsetTime.innerHTML.substring( 0, 2 );
    m = sunsetTime.innerHTML.substring( 3, 5 );
    sunset = ( h * 60 ) + ( m * 1 );
    
    // Now that we're past the code that is predicated on 24 hour clock format, see if we're displaying 12 hour time.
    
    if ( ! use24HourClock ) {
        sunriseTime.textContent = clock24To12( sunriseTime.innerHTML );
        sunsetTime.textContent = clock24To12( sunsetTime.innerHTML );
    }
    
    // Mark solar noon, nadir, dawn, dusk, and length of twilight.
    
    if ( ( times.nauticalDawn instanceof Date ) && ( ! isNaN( times.nauticalDawn) ) ) {
        dawn = ( times.nauticalDawn.getHours() * 60 ) + ( times.nauticalDawn.getMinutes() * 1 );
        dusk = ( times.nauticalDusk.getHours() * 60 ) + ( times.nauticalDusk.getMinutes() * 1 );
    } else {
        dawn = ( times.dawn.getHours() * 60 ) + ( times.dawn.getMinutes() * 1 );
        dusk = ( times.dusk.getHours() * 60 ) + ( times.dusk.getMinutes() * 1 );
    }
    twiLen = sunrise - dawn;
    
    solarNoon = ( times.solarNoon.getHours() * 60 ) + ( times.solarNoon.getMinutes() * 1 );
    solarMidnight = ( times.nadir.getHours() * 60 ) + ( times.nadir.getMinutes() * 1 );
    
    var currentDayLength = ( times.sunset.getHours()  * 3600 + times.sunset.getMinutes()  * 60 + times.sunset.getSeconds()  ) -
                           ( times.sunrise.getHours() * 3600 + times.sunrise.getMinutes() * 60 + times.sunrise.getSeconds() );
    
    // Must recompute previousDayLength as Today can change.
    
    previousDayLength = pdl();
    
    trendingDayLength = currentDayLength - previousDayLength;

    // Luna.
    
    var moonTimes = SunCalc.getMoonTimes( now, latitude, longitude );
    if ( moonTimes.alwaysUp ) {
        moonrise = 0;
        moonset = mpd;
    } else if ( moonTimes.alwaysDown ) {
        moonrise = -1;
        moonset = -1;
    } else if ( moonTimes.rise && moonTimes.set ) {
        moonrise = ( moonTimes.rise.getHours() * 60 ) +  ( moonTimes.rise.getMinutes() * 1 ) ;
        moonset = ( moonTimes.set.getHours() * 60 ) +  ( moonTimes.set.getMinutes() * 1 ) ;
    } else if ( ! moonTimes.rise ) {
        moonset = ( moonTimes.set.getHours() * 60 ) +  ( moonTimes.set.getMinutes() * 1 ) ;
        moonrise = -1;
    } else if ( ! moonTimes.set ) {
        moonrise = ( moonTimes.rise.getHours() * 60 ) +  ( moonTimes.rise.getMinutes() * 1 ) ;
        moonset = -1;
    } else {
        console.log("hmm, what to do now??");
    }
    
    moonriseTime.textContent = moonrise == -1 ? "" : addZero( moonTimes.rise.getHours() ) + ':' + addZero( moonTimes.rise.getMinutes() );
    moonsetTime.textContent  = moonset  == -1 ? "" : addZero( moonTimes.set.getHours() )  + ':' + addZero( moonTimes.set.getMinutes() );

    if ( ! use24HourClock ) {
        moonriseTime.textContent = clock24To12( moonriseTime.innerHTML );
        moonsetTime.textContent = clock24To12( moonsetTime.innerHTML );
    }

    var trend;
    
    // Detailed info, Sol.

    var extraSpaces = use24HourClock ? "\u00A0\u00A0" : "";
    document.getElementById( "detailedInfo2" ).textContent = "Dawn: " + sqzSpc( dayMinute2HumanTime( dawn ) ) + extraSpaces + "\u00A0" + 'Dusk: ' + sqzSpc( dayMinute2HumanTime( dusk ) );
    document.getElementById( "detailedInfo3" ).textContent = "Noon: " + sqzSpc( dayMinute2HumanTime( solarNoon ) ) + extraSpaces + "\u00A0" + 'Midn: ' + sqzSpc( dayMinute2HumanTime( solarMidnight ) );
    var solarDist;
    [solarDist, trend] = Sol.earthDistance( now );
    var currentSolarDist = parseFloat( solarDist ).toFixed( 0 );
    var sunPos = SunCalc.getPosition( now, latitude, longitude);
    var solarAltitude = sunPos.altitude;
    var solarAltitudeDegrees = solarAltitude * ( 180 / Math.PI );
    var solarAzimuth = sunPos.azimuth; // measured from south, moving west !!!!
    solarAzimuth -= Math.PI;		// now normalize so we start from the north !
    if ( solarAzimuth < 0 ) {
        solarAzimuth += 2 * Math.PI;
    }
    var solarAzimuthDegrees = solarAzimuth * ( 180 / Math.PI );
    document.getElementById( "detailedInfo4" ).textContent = "Alti: " + fixDegrees( parseFloat( solarAltitudeDegrees ).toFixed( 0 ) ) + "\xB0" + "\u00A0\u00A0\u00A0\u00A0" + 'Azim: ' + fixDegrees( parseFloat( solarAzimuthDegrees ).toFixed( 0 ) ) + "\xB0";
    document.getElementById( "detailedInfo5" ).textContent = "Dist: " + commify( currentSolarDist ) + " km" + trend;
    var v;
    [v, trend] = Sol.earthOrbitalSpeed( solarDist ); // km/dy
    v /= 24;
    
    document.getElementById( "detailedInfo5.1" ).textContent = "Velo: " + "\u00A0\u00A0\u00A0\u00A0" + commify( parseFloat( v ).toFixed(0) ) + " km/h" + trend;

    // Detailed info, Luna.

    var moonPos = SunCalc.getMoonPosition( now, latitude, longitude);
    document.getElementById( "detailedInfo7" ).textContent = "Rise: " + sqzSpc( dayMinute2HumanTime( moonrise ) ) + extraSpaces + "\u00A0" + ' Set: ' + sqzSpc( dayMinute2HumanTime( moonset ) );
    var lunarAltitude = moonPos.altitude;
    var lunarAltitudeDegrees = lunarAltitude * ( 180 / Math.PI );
    var lunarAzimuth = moonPos.azimuth; // measured from south, moving west !!!!
    lunarAzimuth -= Math.PI;		// now normalize so we start from the north !
    if ( lunarAzimuth < 0 ) {
        lunarAzimuth += 2 * Math.PI;
    }
    lunarAzimuthDegrees = lunarAzimuth * ( 180 / Math.PI );
    document.getElementById( "detailedInfo8" ).textContent = "Alti: " + fixDegrees( parseFloat( lunarAltitudeDegrees ).toFixed( 0 ) ) + "\xB0" + "\u00A0\u00A0\u00A0\u00A0" + 'Azim: ' + fixDegrees( parseFloat( lunarAzimuthDegrees ).toFixed( 0 ) ) + "\xB0";
    var lunarDist;
    [lunarDist, trend] = Luna.earthDistance( now );
    var currentLunarDist = parseFloat( lunarDist ).toFixed( 0 );
    extraSpaces = "\u00A0\u00A0\u00A0\u00A0";
    document.getElementById( "detailedInfo9" ).textContent = "Dist: " + extraSpaces + commify( currentLunarDist ) + " km" + trend;

    [v, trend] = Luna.earthOrbitalSpeed( lunarDist );
    v /= 24;
    document.getElementById( "detailedInfo9.1" ).textContent = "Velo: "  + "\u00A0\u00A0\u00A0\u00A0\u00A0\u00A0" + commify( parseFloat( v ).toFixed(0) ) + " km/h" + trend;
    document.getElementById( "detailedInfo9.2" ).textContent = "Phas: " + Luna.phase( now );

    // Introducing a horizontal artificial horizon causes a skew of ( sunrise - halfNightMinutes ) minutes
    // we must adjust for, so that sunrise and sunset on the solar sine path intersect the horizon
    // horizontally.
    //
    // Solar noon is the moment that divides the time interval from sunrise to sunset exactly in half.
    // var nnn = solarNoon - (12*60); // Hmmmm, nnn is the same as solarSkew ?
    
    var dayMinutes = sunset - sunrise;
    var halfNightMinutes = ( mpd - dayMinutes ) / 2;
    solarSkew = sunrise - halfNightMinutes;
    solarSkew = solarSkew.toFixed( 0 );

    // Define a list of special minutes that we can simulate rapidly.
    
    specialMinutes = [];
    specialMinutes.push( dawn );
    specialMinutes.push( sunrise );
    specialMinutes.push( mpd / 2 ); // noon
    specialMinutes.push( solarNoon );
    specialMinutes.push( sunset );
    specialMinutes.push( dusk );
    specialMinutes.push( mpd - 1 ); // midnight
    specialMinutes.push( solarMidnight );
    specialMinutes.push( getNowMinute() ); //Now
    if ( moonrise != -1 ) {
        specialMinutes.push( moonrise );
    }
    if (  moonset != -1 ) {
        specialMinutes.push( moonset );
    }
    specialMinutes.sort( function( a, b ) { return a - b } );

    //checkForSolarEclipseHide( now, solarAltitudeDegrees, solarAzimuthDegrees, lunarAltitudeDegrees, lunarAzimuthDegrees);
    
} // end updateRiseAndSetTimes

// MARK: - New day

function newDay ( minute ) {

    // Start new day animation.
    
    var SVGContainer = document.getElementById( 'solarcoaster' );
    var content = SVGContainer.innerHTML;
    SVGContainer.innerHTML = content;
    init1();
    if ( ! screenSaver ) {
        var oldSlider = document.getElementById( 'slider-SolarCoaster' );
        oldSlider.innerHTML = "";
        makeSlider();
        moveSlider( minute );
        var newSlider = document.getElementById( 'slider-SolarCoaster' );
        newSlider.setAttribute( 'visibility', 'hidden' );
    }

    newDayInProgress = 1;
    setMarkersVisibility( 'hidden' );
    hideAllLunarElements();
    document.getElementById( 'slider-SolarCoaster' ).setAttribute( 'visibility', 'hidden' );
    document.getElementById( 'youDriveSpecialMinutesElement' ).setAttribute( 'visibility', 'hidden' );
    document.getElementById( 'youDriveAllMinutesElement' ).setAttribute( 'visibility', 'hidden' );
    document.getElementById( 'eclipse1' ).setAttribute( 'visibility', 'hidden' );
    document.getElementById( 'eclipse2' ).setAttribute( 'visibility', 'hidden' );
    document.getElementById( 'eclipse3' ).setAttribute( 'visibility', 'hidden' );
    document.getElementById( 'eclipse4' ).setAttribute( 'visibility', 'hidden' );

    var svg = document.getElementById( 'solarSineElement' );
    svg.setCurrentTime( 0 );
    
    var ssp = document.getElementById( 'solarSinePath' );
    var sspc = document.getElementById( 'solarSinePathClone' );
    var horizonWide = document.getElementById( 'solarHorizonWide' );
    
    sspc.setAttribute( 'visibility', 'visible' );
    horizonWide.setAttribute( 'visibility', 'visible' );
    
    var dur = '1.5s';
    
    var ani = document.createElementNS("http://www.w3.org/2000/svg","animateTransform");
    ani.setAttribute("attributeName", "transform");
    ani.setAttribute("attributeType", "xml");
    ani.setAttribute("id", "newDay" );
    ani.setAttribute("type", "translate" );
    ani.setAttribute("from", "0 0");
    ani.setAttribute("to", "-925 0");
    ani.setAttribute("begin", '0s');
    ani.setAttribute("dur", dur);
    ssp.appendChild( ani ); // animate solar sine path
    
    ani = document.createElementNS("http://www.w3.org/2000/svg","animateTransform");
    ani.setAttribute("attributeName", "transform");
    ani.setAttribute("attributeType", "xml");
    ani.setAttribute("id", "newDay2" );
    ani.setAttribute("type", "translate" );
    ani.setAttribute("from", "925 0"); // 910
    ani.setAttribute("to", "0 0");
    ani.setAttribute("dur", dur);
    sspc.appendChild( ani ); // animate solar sine path clone
    
    var endAnim = "endAnimate( " + minute + " )";
    ani = document.createElementNS("http://www.w3.org/2000/svg","animateTransform");
    ani.setAttribute("attributeName", "transform");
    ani.setAttribute("attributeType", "xml");
    ani.setAttribute("id", "newDay3" );
    ani.setAttribute("type", "translate" );
    ani.setAttribute("from", "861 849");
    ani.setAttribute("to", "-61 849");
    ani.setAttribute("begin", '0s');
    ani.setAttribute("dur", dur);
    ani.setAttribute( 'onend', endAnim );
    sol.appendChild( ani ); // animate Sol

} // end newDay

function endAnimate( minute ) {
    
    // Finish new day animation.
    
    var ssp = document.getElementById( 'solarSinePath' );
    var sspc = document.getElementById( 'solarSinePathClone' );
    //var sol = document.getElementById( 'solarSol' );
    var horizonWide = document.getElementById( 'solarHorizonWide' );
    
    sol.setAttribute( 'transform', 'translate( 51, 899 )' );
    sspc.setAttribute( 'visibility', 'hidden' );
    horizonWide.setAttribute( 'visibility', 'hidden' );
    
    var animation = document.getElementById('newDay');
    var parent = animation.parentNode;
    parent.removeChild(animation);
    
    animation = document.getElementById('newDay2');
    parent = animation.parentNode;
    parent.removeChild(animation);
    
    animation = document.getElementById('newDay3');
    parent = animation.parentNode;
    parent.removeChild(animation);
    
    var vis =  showDetails ? 'visible' : 'hidden';
    setMarkersVisibility( vis );
    setLunaVis( 'visible' );
    lunaSinePath.setAttribute( 'opacity', 1.0 );
    lunaPath.setAttribute( 'opacity', 1.0 );
    lunabackPath.setAttribute( 'opacity', 1.0 );
    document.getElementById( 'slider-SolarCoaster' ).setAttribute( 'visibility', 'visible' );
    document.getElementById( 'youDriveSpecialMinutesElement' ).setAttribute( 'visibility', 'visible' );
    document.getElementById( 'youDriveAllMinutesElement' ).setAttribute( 'visibility', 'visible' );
    document.getElementById( 'eclipse1' ).setAttribute( 'visibility', 'visible' );
    document.getElementById( 'eclipse2' ).setAttribute( 'visibility', 'visible' );
    document.getElementById( 'eclipse3' ).setAttribute( 'visibility', 'visible' );
    document.getElementById( 'eclipse4' ).setAttribute( 'visibility', 'visible' );
    newDayInProgress = 0;

    animateBodies( minute );

} // end endAnimate

// MARK: - Onclick

function dismissModal() {
    
    document.getElementById('myModalWindow').style.display = "none";
    document.getElementById('modal-content-here').style['background-color'] = '#cacaca'; // at this time do not how to save orginal color
    myAlertActive = 0;

}

function displayInfo ( ) { // iOS only
    
    if ( iOSInfoMsg != '' ) {
        var help = document.getElementById( 'modal-content-here' );
        var darkModeBackgroundColor = (darkMode == "dark" ? 'black' : '#03a6f3');
        help.style['background-color'] = ( backgroundColor == 'auto' ? darkModeBackgroundColor : backgroundColor );
        help.style['color'] = 'white';
        myAlert( '', iOSInfoMsg );
    }
    
} // end displayInfo

function onclickFunction( evt ) {

    evt.stopPropagation();
    
    if ( screenSaver ) {
        var o = screenSaver % 3; // 0 showOnlyMarkers(); 1 youDrive1; 2 youDrive2
        screenSaver++;
        var version = document.getElementById( 'version' )
        showDetails = 1;
        version.setAttribute( 'visibility', 'visible' );
        if ( o == 0 ) {
            solarSineElement( 'SS0' );
            version.setAttribute( 'visibility', 'hidden' );
            showOnlyMarkers();
            setDetailedInfoVisibility( 'visible' );
        } else if ( o == 1 ) {
            setDetailedInfoVisibility( 'visible' );
            youDriveSpecialMinutesElement( 'SS1' );
        } else if ( o == 2 ) {
            youDriveAllMinutesElement( 'SS2' );
        }
        return;
    }
    
    if ( todayExt || newDayInProgress ) {
        return;
    }
    
    var id = evt.target.id;
    if ( event.target == document.getElementById('modal-content-here') || event.target == document.getElementById('myModalWindow') || event.target == document.getElementById('myModalOK') ) {
        dismissModal();
        return;
    }

    if ( id == 'triangleDisclosure' ) {
        if ( showDetails ) {
            document.getElementById("ani-triangleDisclosure-anticlock").beginElement();
        } else {
            document.getElementById("ani-triangleDisclosure-clockwise").beginElement();
        }
        solarSineElement( id );
    } else if ( id == 'youDriveSpecialMinutesElement' ) {
        youDriveSpecialMinutesElement( id );
    } else if ( id == 'youDriveAllMinutesElement' ) {
        youDriveAllMinutesElement( id );
    } else if ( id == 'solarSol' || id == 'solarSolHalo' ) {
        sunElement( id );
    } else if ( id == 'eclipse1' || id == 'eclipse2' ) {
        eclipseElement( id );
    } else if ( id == 'eclipse3' || id == 'eclipse4' ) {
        lunarEclipseElement( id );
    } else if ( id == 'luna' || id == 'lunaBack' ) {
        var opac1 = lunaPath.getAttribute( 'opacity' );
        var opac2 = lunabackPath.getAttribute( 'opacity' );
        if ( opac1 == 1 && opac2 == 1 ) {
            moonElement( id );
        }
    } else if ( id == 'triangleUp' || id == 'triangleDn' || id == 'solarHorizon' ) {
        triangleElement( id );
    } else if ( id == 'help'  || id == 'helpCircle') {
        displayInfo();
    } else  {
        var markers = {
            "dawnCircle:Nautical Dawn"           : dawn,
            "duskCircle:Nautical Dusk"           : dusk,
            "solarNoonCircle:Solar Noon"         : solarNoon,
            "noonCircle:Noon"                    : ( mpd / 2 ),
            "midnightCircle:Midnight"            : 0,
            "solarMidnightCircle:Solar Midnight" : solarMidnight,
            "solarMidnightCircleLeft:Solar Midnight" : solarMidnight,
            "solarMidnightCircleRight:Solar Midnight" : solarMidnight,
            "sunriseCircle:Sunrise"              : sunrise,
            "sunsetCircle:Sunset"                : sunset,
            "moonriseCircle:Moonrise"            : moonrise,
            "moonsetCircle:Moonset"              : moonset,
            "nowCircle:Now"                      : getNowMinute(),
        };
        var circle;
        for ( var key in markers ) {
            if ( markers.hasOwnProperty( key ) ) {
                var toks = key.split( ':' );
                if ( id == toks[0] && markers[ key ] != -1 ) {
                    alert( toks[1] + ': ' + dayMinute2HumanTime( markers[ key ] ) );
                    return;
                }
            }
        }
    }
    
} // end onclickFunction

function showOnlyMarkers () {
    
    setMarkersVisibility( 'visible' );
    title.setAttribute( 'visibility', 'visible' );
    subtitle.setAttribute( 'visibility', 'visible' );
    titleTime.setAttribute( 'visibility', 'hidden' );
    
}

function solarSineElement ( id ) {
    
    showDetails = 1 - showDetails;
    var vis = showDetails ? 'visible' : 'hidden';
    setMarkersVisibility( vis );
    setDetailedInfoVisibility( vis );
    title.setAttribute( 'visibility', vis );
    subtitle.setAttribute( 'visibility', vis );
    titleTime.setAttribute( 'visibility', vis );
    
    var youDrive1 = document.getElementById( "youDriveSpecialMinutesElement" );
    var youDrive2 = document.getElementById( "youDriveAllMinutesElement" );
    var version = document.getElementById( 'version' );
    youDrive1.setAttribute( "visibility", vis );
    youDrive2.setAttribute( "visibility", vis );
    var slider = document.getElementById( 'slider-SolarCoaster' );
    slider.setAttribute( 'visibility', vis );
    version.setAttribute( 'visibility', vis );
    document.getElementById( 'eclipse1' ).setAttribute( 'visibility', vis );
    document.getElementById( 'eclipse2' ).setAttribute( 'visibility', vis );
    document.getElementById( 'eclipse3' ).setAttribute( 'visibility', vis );
    document.getElementById( 'eclipse4' ).setAttribute( 'visibility', vis );
    var helpVis = showDetails ? 'visible' : 'visible';
    if ( iOSInfoMsg == '' ) {
        helpVis = 'hidden';
    }
    document.getElementById( 'help' ).setAttribute( 'visibility', helpVis );
    document.getElementById( 'helpCircle' ).setAttribute( 'visibility', helpVis );

    clearTimeout( timerHandle );
    var minute = getNowMinute();
    if ( vis == 'hidden' ) {
        youDrive = youDriveNone;
        initTrendingData();
        debugDate = debugDateOrig;
        animateBodies( solarMidnight + 1 );
    } else {
        moveSlider( minute );
        timerHandle = setTimeout( function() { animateDetails() }, 1000 * 60 * minutesInterval );
    }
    
} // end solarSineElement

function youDriveSpecialMinutesElement( id ) {
    
    if ( youDrive == youDriveAllMinutes ) {
        youDriveAllMinutesElement ();
    }
    youDrive = ( youDrive != youDriveNone ? youDriveNone : youDriveSpecialMinutes );
    clearTimeout( timerHandle );
    if ( youDrive ) {
        animateBodies( 0 ); // specialMinutes index 0
    } else {
        moveSlider( sliderObj.youDriveMinute );
        animateBodies( sliderObj.youDriveMinute ); // too bad it always goes to Now minute
    }
    
} // end youDriveSpecialMinutesElement

function youDriveAllMinutesElement( id ) {
    
    if ( youDrive == youDriveSpecialMinutes ) {
        youDriveSpecialMinutesElement ();
    }
    youDrive = ( youDrive != youDriveNone ? youDriveNone : youDriveAllMinutes );
    clearTimeout( timerHandle );
    if ( youDrive ) {
        animateBodies( solarMidnight + 1 );
    } else {
        moveSlider( sliderObj.youDriveMinute );
        animateBodies( sliderObj.youDriveMinute ); // too bad it always goes to Now minute
    }

} // end youDriveAllMinutesElement

// MARK: - SVG slider

function moveSlider ( minute ) {
    
    // Take a minute of the day amd move the slider correspondingly.
    
    var pc = minute / mpd;    // percent, 0 <= pc <= 1
    pc = pc.toFixed(8)
    
    
    pc = 1 - pc;
    var sv = pc * mpd;
    sliderObj.value = sv;
    sliderObj.youDriveMinute = minute;
    
} // end moveSlider

function makeSlider () {
    
    sliderObj = new svgSlider(document.getElementById('slider-SolarCoaster'), {
        markers: false,
        palette: 'light',
        cursor_color: "yellow",
        cursor_width: 10,
        cursor_height: 10,
        cursor_radius: 10,
        track_color: "white",
        track_bg_color: 'white',
        track_bg_border_color: 'white',
        track_bg_border_width: 0.4,
        track_length: 90,
        track_width: 0.4,
        track_offset: 5,
        track_bg_width: 0.4,
        value_max: mpd - 1,
        value_min: 0,
        value_resolution: null,
    });
    sliderObj.youDriveMinute = 0

    var sliderElem = document.getElementById('slider-SolarCoaster');
    //var sizes = [ 2.3, 2050, 2100 ]; // larger scale means smaller slider: [ scale, width, height ]
    var sizes = [ 1.4, 1800, 1000 ]; // larger scale means smaller slider: [ scale, width, height ] - this is really just trial and error
    sliderElem.setAttribute( "viewBox", "0 " + -0.4*sizes[0] + " " + 20*sizes[0] + " " + 100*sizes[0]);
    sliderElem.setAttribute( 'width', sizes[1] );
    sliderElem.setAttribute( 'height', sizes[2] );
    sliderElem.setAttribute( 'visibility', 'hidden' );

    sliderElem.addEventListener("change", function(event) {
        // Event.target: a reference to the object that dispatched the event. It is different from event.currentTarget
        //               when the event handler is called during the bubbling or capturing phase of the event.
        var pc = `${event.detail}` / mpd;
        pc = 1.0 - pc; // for consistency with previous g3 slider
        var minute = parseFloat( ( pc * ( mpd ) ) ) + parseFloat( solarSkew );
        updateRiseAndSetTimes ( minute ); // update Sol and Luna rise and set times, minute of the day
        animateSol ( minute );
        animateLuna ( minute );
    });

    moveSlider( 0 );

} // end makeSlider

// MARK: - Mostly called from Objective-C wrapper

function deadWebviewCheck () { // always return ALIVE indicating we are alive
    
    // Obj-C wrapper calls us to ensure Solar Coaster JS is still running.  On failure
    // the wrapper code will make a new WkWebView instance and reload us.  A special
    // case of the White Screen Of Death (WSOD).
    
    return 'ALIVE';
    
} // end deadWebviewCheck

function imminentEclipseCircumstances () {
    
    var eclipseHTML = ' ';
    if ( imminentEclipse ) {
        if ( imminentEclipse.length == 16 ) { // solar eclipse
            solarEclipses = [];
            solarEclipses.push( imminentEclipse );
            eclipseHTML = showSolarEclipseData( 1 );
            debugDate = '';
        } else if ( imminentEclipse.length == 18 ) { // lunar eclipse
            lunarEclipses = [];
            lunarEclipses.push( imminentEclipse );
            eclipseHTML = showLunarEclipseData( 1 );
            debugDate = '';
        }
    }
    return eclipseHTML;
    
} // end imminentEclipseCircumstances

function injectiOSInfoMsg ( msg ) {
    
    iOSInfoMsg = msg;
    document.getElementById( 'help' ).setAttribute( 'visibility', ( iOSInfoMsg == '' || todayExt ) ? 'hidden' : 'visible' );
    document.getElementById( 'helpCircle' ).setAttribute( 'visibility', ( iOSInfoMsg == '' || todayExt ) ? 'hidden' : 'visible' );

} // end injectiOSInfoMsg

function onceDailyEclipseCheck () { // Apps only
    
    // Collect all eclipses, sort by time, and return summary for Apps. Remember imminent eclipse for the screen saver.
    
    if ( todayExt == 0 ) {
        
        var eclipseSummary = [];
        var allEclipses = [];
        imminentEclipse = [];
        
        solarEclipses = [];
        setSolartimeperiod( "SE2001" );
        var numSolarEclipses = solarEclipses.length;
        if ( numSolarEclipses > 0 ) {
            for ( var i = 0; i < numSolarEclipses; i++ ) {
                var eventDetails = solarEclipses[ i ]; // array of 16 : 2017-Aug-21, P, 11:38:05, 52, 133, -, 13:03:49, 60, 168, -, 14:29:57, 58, 209, 0.985, 0.987, -
                allEclipses.push(eventDetails);
            }
        }
        
        lunarEclipses = [];
        setLunartimeperiod( "LE2001" );
        var numLunarEclipses = lunarEclipses.length;
        console.log('numLuanrEclipses=',numLunarEclipses);
        if ( numLunarEclipses > 0 ) {
            for ( var i = 0; i < numLunarEclipses; i++ ) {
                var eventDetails = lunarEclipses[ i ]; // array of 18 : 2018-Jan-31, T, 2.294,  1.315, 10:51, +29, 11:48, +19, 12:52, +07, 13:30, +01, italic-14:08, italic--06, italic-15:11, italic--15, italic-16:08, italic--23
                console.log(eventDetails);
                allEclipses.push( eventDetails );
            }
        }
        
        allEclipses.sort( function(a,b) {
         var fullMonth2 = {
         'Jan' : '01',
         'Feb' : '02',
         'Mar' : '03',
         'Apr' : '04',
         'May' : '05',
         'Jun' : '06',
         'Jul' : '07',
         'Aug' : '08',
         'Sep' : '09',
         'Oct' : '10',
         'Nov' : '11',
         'Dec' : '12',
         };
         
         var aDate = a[0];
         var y = aDate.substring(0, 4);
         var m = aDate.substring(5, 8);
         var d = aDate.substring(9, 11);
         m = fullMonth2[ m ];
         aDate = y + '-' + m + '-' + d;
         var aEpoch = new Date( aDate ).getTime() / 1000;
         
         var bDate = b[0];
         var y = bDate.substring(0, 4);
         var m = bDate.substring(5, 8);
         var d = bDate.substring(9, 11);
         m = fullMonth2[ m ];
         bDate = y + '-' + m + '-' + d;
         var bEpoch = new Date( bDate ).getTime() / 1000;
         
         return aEpoch - bEpoch;
         });
        
        imminentEclipse = allEclipses[0];
        console.log('imE=', imminentEclipse,  ', len=',imminentEclipse.length);
        
        var numEclipses = allEclipses.length;
        if ( numEclipses > 0 ) {
            var event;
            for ( var i = 0; i < numEclipses; i++ ) {
                var eventDetails = allEclipses[ i ]; // array of 16 or 18
                event =  eTypeSolar[ eventDetails[1] ] + ' solar eclipse '; // assume solar eclipse
                if ( eventDetails.length == 18 ) {
                    event = eType[ eventDetails[1] ] + ' lunar eclipse ';
                }
                eclipseSummary.push( event + lunarHuman2solarHumanDate( eventDetails[0] ) );
            }
        }
        console.log('ES=',eclipseSummary);
        return eclipseSummary;
        
    } // if not todayExt

} // end onceDailyEclipseCheck

function setBackground ( color ) {
                
    backgroundColor = color;
    if ( backgroundColor == 'auto' ) {
        color = ( darkMode == 'dark' ? '#000000' : '#03a6f3' );
    }
    var element    = document.getElementById( 'solarcoaster' );
    var help       = document.getElementById( 'help' );
    var helpCircle = document.getElementById( 'helpCircle' );
    element.style['background'] = color;
    element = document.getElementById( 'bodyid' );
    element.style['background'] = color;
    element = document.getElementById( 'version' );
    if ( color == '#03a6f3' ) {
        element.setAttribute( 'fill', 'blue' );
        help.setAttribute( 'fill', 'blue' );
        helpCircle.setAttribute( 'fill', 'blue' );
    } else {
        element.setAttribute( 'fill', '#03a6f3' );
        help.setAttribute( 'fill', '#03a6f3' );
        helpCircle.setAttribute( 'fill', '#03a6f3' );
    }
    return 0;
    
} // end setBackground

function setLatLonAlt ( lat, lon, alt ) {
    
    latitude = lat;
    longitude = lon;
    altitude = alt;
    clearTimeout( timerHandle );
    animateBodies( getNowMinute() );

} // end setLatLonAlt

function setLunaVis ( vis ) {
    
    // Times are always hidden in v4.0 and greater.
    var vis = 'hidden';
    moonriseTime.setAttribute( 'visibility', vis );
    moonsetTime.setAttribute( 'visibility', vis );
    
} // end setLunaVis

function setScreenSaverSize ( height ) {
    
    screenSaver = 1;
    showOnlyMarkers();
    showDetails = 1;
    setDetailedInfoVisibility( 'visible' );
    
    var w = height * ( 140 / 258 );            // keep same aspect ratio as HTML, see GlobalConst.h
    var h = height;
    var element = document.getElementById( 'solarcoaster' );
    element.style['width'] = w + 'px';
    element.style['height'] = h + 'px';
    var disclosure = document.getElementById('triangleDisclosure');
    disclosure.setAttribute( 'visibility', 'hidden' );

} // end setScreenSaverSize

function setTimeFormat ( tf ) {    // 0 = 12 hour, 1 = 24 hour
    
    use24HourClock = tf;
    updateRiseAndSetTimes( getNowMinute() );    // repaint rise and set times
    updateTitle( getNowMinute() );
    return 0;
    
} // end setTimeFormat

function setTodayExt ( ext ) {
    
    todayExt = ext;
    
    var element;
    if ( todayExt ) {
        element = document.getElementById( 'solarcoaster' );
        element.style['width'] = '510px';
        element.style['height'] = '960px';
        element = document.getElementById( 'lunaSinePath' );
        element.setAttribute( 'stroke-width', '10px' );
        element = document.getElementById( 'solarSinePath' );
        element.setAttribute( 'stroke-width', '10px' );
        element = document.getElementById( 'solarSinePathClone' );
        element.setAttribute( 'stroke-width', '10px' );
        showOnlyMarkers();
        showDetails = 1;
        setDetailedInfoVisibility( 'visible' );
        document.getElementById( 'help' ).setAttribute( 'visibility', 'hidden' );
        document.getElementById( 'helpCircle' ).setAttribute( 'visibility', 'hidden' );
        var disclosure = document.getElementById('triangleDisclosure');
        disclosure.setAttribute( 'visibility', 'hidden' );
    } else {
        element = document.getElementById( 'solarcoaster' );
        element.style['width'] = 'auto';
        element.style['height'] = 'auto';
        element = document.getElementById( 'lunaSinePath' );
        element.setAttribute( 'stroke-width', '6px' );
        element = document.getElementById( 'solarSinePath' );
        element.setAttribute( 'stroke-width', '6px' );
        element = document.getElementById( 'solarSinePathClone' );
        element.setAttribute( 'stroke-width', '6px' );
    }
    
    return 0;
    
} // end setTodayExt

function setVersion ( v ) {
    
    version = v;
    document.getElementById( 'version' ).textContent = 'Solar Coaster'; // make updating docs easier without version ' +  version;
    return 0;
    
} // end setVersion

// MARK: - Miscellaneous

function pdl () { // previousDayLength
    
    var yesterday = getNowDateObject();
    yesterday.setDate( yesterday.getDate() - 1 );
    var solarTimes = SunCalc.getTimes( yesterday, latitude, longitude, altitude );
    return ( solarTimes.sunset.getHours()  * 3600 + solarTimes.sunset.getMinutes()  * 60 + solarTimes.sunset.getSeconds()  ) -
    ( solarTimes.sunrise.getHours() * 3600 + solarTimes.sunrise.getMinutes() * 60 + solarTimes.sunrise.getSeconds() );

} // end pdl

function sqzSpc ( str ) {
    
    if ( str.length > 7 ) {
        str = str.replace( ' ', '' );
    }
    return str;
    
} // end sqzSpc

function fixDegrees( num ) {

    if ( num == -0 ) {
	num = 0;
    }
    var sign = num >= 0 ? 1 : 0;
    if ( ! sign ) {
	num = -num;
    }
    if ( num < 10 ) {
	if ( sign ) {
	    num = "\u00A0\u00A0" + num;
	} else {
	    num = "\u00A0-" + num;
	}
    } else {
	if ( sign ) {
	    num = "\u00A0" + num;
	} else {
	    num = '-' + num;
	}
    }
    return num;

} // end fixDegrees

//
// Sol functions.
//

function gradientFillSol ( minute ) {
    
    if ( minute >= 0 && minute < dawn ) {
        isNight( minute );
    } else if ( ( minute > dawn ) && ( minute <= dawn + twiLen ) ) {
        isMorningTwilight( minute );
    } else if ( ( minute >= dawn + twiLen ) && ( minute < dusk - twiLen ) ) {
        isDay( minute );
    } else if ( ( minute >= dusk - twiLen ) && ( minute < dusk ) ) {
        isEveningTwilight( minute );
    } else if ( ( minute >= dusk ) ) {
        isNight( minute );
    }
    
} // end gradientfillSol

function updateTitle ( minute ) {
    
    var now = getNowMinute();
    if ( minute >= dawn-1 && minute <= dawn + ( minutesInterval - 1 )  ) {
        newTitle( "DAWN", minute, dayMinute2HumanTime( dawn ) );
    } else if ( minute >= sunrise-1 && minute <= sunrise + ( minutesInterval - 1 ) ) {
        newTitle( 'SUNRISE', minute, dayMinute2HumanTime( sunrise ) );
    } else if ( minute >= ( mpd / 2 )-1 && minute <= ( mpd / 2 ) + ( minutesInterval - 1 ) ) {
        newTitle( 'NOON', minute, dayMinute2HumanTime( mpd / 2 ) );
    } else if ( minute >= solarNoon-1 && minute <= solarNoon + ( minutesInterval - 1 ) ) {
        newTitle( 'SOLAR NOON', minute, dayMinute2HumanTime( solarNoon ) );
    } else if ( minute >= sunset-1 && minute <= sunset + ( minutesInterval - 1 ) ) {
        newTitle( 'SUNSET', minute, dayMinute2HumanTime( sunset ) );
    } else if ( minute >= dusk-1 && minute <= dusk + ( minutesInterval - 1 )  ) {
        newTitle( "DUSK", minute, dayMinute2HumanTime( dusk ) );
    } else if ( minute >= ( mpd - 1 )-1 && minute <= ( mpd - 1 ) + ( minutesInterval - 1 )  ) {
        newTitle( "MIDNIGHT", minute, dayMinute2HumanTime( 0 ) );
    } else if ( minute >= solarMidnight-1 && minute <= solarMidnight + ( minutesInterval - 1 )  ) {
        newTitle( "SOLAR MIDNIGHT", minute, dayMinute2HumanTime( solarMidnight ) );
    } else if ( minute >= now-1 && minute <= now + ( minutesInterval - 1 ) ) {
        newTitle( 'NOW', minute, dayMinute2HumanTime( now ) );
    } else if ( minute >= moonrise-1 && minute <= moonrise + ( minutesInterval - 1 ) && moonrise != -1 ) {
        newTitle( 'MOONRISE', minute, dayMinute2HumanTime( moonrise ) );
    } else if ( minute >= moonset-1 && minute <= moonset + ( minutesInterval - 1 ) &&  moonset != -1 ) {
        newTitle( 'MOONSET', minute, dayMinute2HumanTime( moonset ) );
    }

} // end updateTitle

function isMorningTwilight ( minute ) {
    
    newTitle( "MORNING TWILIGHT", minute, dayMinute2HumanTime( minute ) );
    
    var m = minute - dawn; // minutes 1, 2, 3, ... 60
    var x = twiLen / 2;
    
    // Brighten Sol, and tighten and fadein the halo.
    //
    // The radial gradient controls the color of under-Sol. For the first x minutes of twilight the color varies
    // from, your pick, either black (0,0,0) or purple (128,0,128) to red (255,0,0). The last x minutes the color
    // varies from red (255,0,0) to yellow (255,255,0).  stop1 goes from 0% to 70%, stop2 is fixed at 100%.
    
    var stop1 =  document.getElementById("solarGradientStop1");
    var stop2 =  document.getElementById("solarGradientStop2");
    var p = ( ( twiLen - m ) / twiLen ) * 70; // stop p = 0 .. 70, percent - Note: also used for blur dispersion below
    stop1.setAttribute( 'offset', p + '%' );
    
    var r;
    var g;
    var b;
    if ( m > x ) {		// last x minutes of twilight
        r = 255;
        g = ( ( m - x ) / x) * 255;
        g = Math.round( g );
        b = 0;
    } else {			// first x minutes of twilight
        if ( nightSolColor == 'purple' ) {
            r = ( ( ( m - 0 ) / x ) * 128 ) + 128;
            r = Math.round( r );
            g = 0;
            b = ( 1 - ( ( m - 0 ) / x ) ) * 128;
            b = Math.round( b );
        } else {		// from black
            r = ( ( m - 0 ) / x ) * 255;
            r = Math.round( r );
            g = 0;
            b = 0;
        }
    }
    var color = rgbToHex( r, g, b );
    stop1.setAttribute( 'stop-color', color );
    stop2.setAttribute( 'offset', '100%' );
    stop2.setAttribute( 'stop-color', 'yellow' );
    
    // The blur filter stdDev controls the diffusness of the over-Sol halo.
    
    solHalo.setAttribute( "visibility", 'visible' );
    var blur = document.getElementById( 'solarBlur' );
    var stdDev;
    stdDev = 82 + ( twiLen - m );
    blur.setAttribute( 'stdDeviation', stdDev);
    p = ( 100 - p ) / 100;	// opacity p = 0.3 .. 1.0
    solHalo.setAttribute( 'opacity', p );
    
} // end isMorningTwilight

function isDay ( minute ) {
    
    newTitle( "DAY", minute, dayMinute2HumanTime( minute ) );
    
    // Sol is now bright, and the halo visible.
    
    var stop1 =  document.getElementById("solarGradientStop1");
    stop1.setAttribute( 'offset', '0%' );
    stop1.setAttribute( 'stop-color', 'yellow' );
    
    var stop2 =  document.getElementById("solarGradientStop2");
    stop2.setAttribute( 'offset', '100%' );
    stop2.setAttribute( 'stop-color', 'yellow' );
    
    solHalo.setAttribute( "visibility", 'visible' );
    var blur = document.getElementById( 'solarBlur' );
    blur.setAttribute( 'stdDeviation', 82 );
    solHalo.setAttribute( 'opacity', 1.0 );
    
} // end isDay

function isEveningTwilight ( minute ) {
    
    newTitle( "EVENING TWILIGHT", minute, dayMinute2HumanTime( minute ) );
    
    var m = dusk - minute; // minutes 60, 59, 58 .... 1
    var x = twiLen / 2;
    
    // Dim Sol, and dispserse and fadeout the halo.
    //
    // The radial gradient controls the color of under-Sol. For the first x minutes of twilight the color varies
    // from yellow (255,255,0) to red (255,0,0). The last x minutes the color varies from red (255,0,0) to, your
    // pick, either black (0,0,0) or purple (128,0,128). stop1 goes from 0% to 70%, stop2 is fixed at 100%.
    
    var stop1 =  document.getElementById("solarGradientStop1");
    var stop2 =  document.getElementById("solarGradientStop2");
    var p = ( ( twiLen - m ) / twiLen ) * 70; // stop p = 0 .. 70, percent - Note: also used for blur dispersion below
    stop1.setAttribute( 'offset', p + '%' );
    
    var r;
    var g;
    var b;
    if ( m > x ) {		// first x minutes of twilight
        r = 255;
        g = ( ( m - x ) / x) * 255;
        g = Math.round( g );
        b = 0;
    } else {			// last x minutes of twilight
        if ( nightSolColor == 'purple' ) {
            r = ( ( ( m - 0 ) / x ) * 128 ) + 128;
            r = Math.round( r );
            g = 0;
            b = ( 1 - ( ( m - 0 ) / x ) ) * 128;
            b = Math.round( b );
        } else {		// to black
            r = ( ( m - 0 ) / x ) * 255;
            r = Math.round( r );
            g = 0;
            b = 0;
        }
    }
    var color = rgbToHex( r, g, b );
    stop1.setAttribute( 'stop-color', color );
    stop2.setAttribute( 'offset', '100%' );
    stop2.setAttribute( 'stop-color', 'yellow' );
    
    // The blur filter stdDev controls the diffusness of the over-Sol halo.
    
    solHalo.setAttribute( "visibility", 'visible' );
    var blur = document.getElementById( 'solarBlur' );
    var stdDev;
    stdDev = 82 + ( twiLen - m );
    blur.setAttribute( 'stdDeviation', stdDev);
    p = ( 100 - p ) / 100;	// opacity p = 1.0 .. 0.3
    solHalo.setAttribute( 'opacity', p );
    
} // end isEveningTwilight

function isNight ( minute ) {
    
    newTitle( "NIGHT", minute, dayMinute2HumanTime( minute ) );
    
    // Sol is now dark, and the halo invisible.
    
    solHalo.setAttribute( "visibility", 'hidden' );
    
    var stop1 =  document.getElementById("solarGradientStop1");
    stop1.setAttribute( 'offset', '70%' );
    stop1.setAttribute( 'stop-color', nightSolColor );
    
    var stop2 =  document.getElementById("solarGradientStop2");
    stop2.setAttribute( 'offset', '100%' );
    stop2.setAttribute( 'stop-color', 'yellow' );
    
} // end isNight

//
// Luna functions.
//

function createPhasedMoon ( minute ) {
    
    // Update the SVG path with the proper arc corresponding to today's phase.  Compute the minute-by-minute
    // phase angle that varies continuously as Luna crosses the sky.  We do this according to SunCalc's note:
    //
    // By subtracting the parallacticAngle from the angle one can get the zenith angle of the moons bright
    // limb (anticlockwise). The zenith angle can be used do draw the moon shape from the observers perspective
    // (e.g. moon lying on its back).
    
    var hhh = Math.floor (minute / 60.0); // hour of day
    var mmm = minute - (hhh*60);	  // minute of day
    var now = getNowDateObject();
    now = new Date( now.getFullYear(), now.getMonth(), now.getDate(), hhh, mmm, 0, 0 );
    var lunaIllumination = SunCalc.getMoonIllumination( now );
    var phase = lunaIllumination.phase;
    var moonPos = SunCalc.getMoonPosition( now, latitude, longitude);
    var radians = lunaIllumination.angle - moonPos.parallacticAngle;
    crescentRotationAngle = radians * ( 180 / Math.PI ); // cresent rotation angle, degrees
    //console.log("illumAngle="+lunaIllumination.angle+", parallacticAngle="+moonPos.parallacticAngle+", radians="+radians+", crescentRotationAngle="+crescentRotationAngle);
    crescentRotationAngle = - crescentRotationAngle+270;
    if ( lunaIllumination.angle > 0 ) {
        crescentRotationAngle = crescentRotationAngle + 180;;
    }
    
    var sweep = [];
    var mag;
    // the "sweep-flag" and the direction of movement change every quarter moon
    // zero and one are both new moon; 0.50 is full moon
    if (phase <= 0.25) {
        sweep = [ 1, 0 ];
        mag = 20 - 20 * phase * 4;
    } else if (phase <= 0.50) {
        sweep = [ 0, 0 ];
        mag = 20 * (phase - 0.25) * 4;
    } else if (phase <= 0.75) {
        sweep = [ 1, 1 ];
        mag = 20 - 20 * (phase - 0.50) * 4;
    } else if (phase <= 1) {
        sweep = [ 0, 1 ];
        mag = 20 * (phase - 0.75) * 4;
    } else {
        console.log("ABRT! minute="+minute);
        exit;
    }
    
    mag /= 2;
    var data = lunaPath.getAttribute( 'd' );
    var d = "m0,0 ";
    d = d + "a" + mag + ",10 0 1," + sweep[0] + " 0,75 ";
    d = d + "a10,10 0 1," + sweep[1] + " 0,-75";
    lunaPath.setAttribute( 'd', d );
    
} // end createPhasedMoon

function setLunaOpacity ( lunaTimeInterval, mins ) { // called IFF Luna is visible
    
    var opacity = 1.0;
    
    // Increase opacity from 0 -> 1 for 2 minutes after moonrise.
    // Decrease opacity from 1 -> 0 for 2 minutes before moonset.
    
    var fadeInStop =   lunaSinePathLen * ( 2 / lunaTimeInterval ) ;
    var fadeOutStart = lunaSinePathLen * ( ( lunaTimeInterval -  2 ) / lunaTimeInterval );
    var now = lunaSinePathLen * (mins / lunaTimeInterval );
    if ( now <= fadeInStop ) {
        opacity = now / fadeInStop;
    }
    if ( now >= fadeOutStart ) {
        opacity = ( lunaSinePathLen -  now ) / ( lunaSinePathLen - fadeOutStart );
    }
    lunaPath.setAttribute( 'opacity', opacity );
    lunabackPath.setAttribute( 'opacity', opacity );
    
} // end setLunaOpacity

//
// Miscellaneous functions.
//

function addZero ( i ) {
    
    j = '' + i;
    if ( j < 10  && j.length < 2 ) {
        j = "0" + j;
    }
    return j;
    
} // end addZero

function clock24To12 (time) {
    
    time = time.toString ().match ( /^([01]\d|2[0-3])(:)([0-5]\d)(:[0-5]\d)?$/ ) || [time];
    
    if (time.length > 1) {	// if time format correct (seconds optional)
        time = time.slice (1);  // remove full string match value
        time[5] = +time[0] < 12 ? ' am' : ' pm';
        time[0] = +time[0] % 12 || 12; // adjust hours
    }
    return time.join ('');  // return adjusted time or original string
}

function componentToHex( c ) {
    
    var d = c;
    if ( d > 255 ) {
        d = 255;
    }
    var hex = d.toString(16);
    return hex.length == 1 ? "0" + hex : hex;
    
} // end componentToHex

function getNowDateObject () { 	// returns a normalized Date object, either now, or a fake now corressponding to the debugDate
    
    var now = new Date();
    if ( debugDate ) {
        var toks = debugDate.split( '-' );
        now = new Date( toks[0], toks[1]-1, toks[2], now.getHours(), now.getMinutes(), 0, 0 );
    }
    return now;
    
} // end getNowDateObject

function getNowMinute () {
    
    var now = getNowDateObject();
    var h = now.getHours();
    var m = now.getMinutes();
    return ( h * 60 ) + ( m * 1 );
    
} // end getNowMinute

function minute2DeltaTime ( minute ) {
    
    var hhh, mmm;
    if ( minute < 0 ) {
        minute = -minute;
    }
    hhh = Math.floor( minute / 60.0 );
    mmm = minute - ( hhh * 60 );
    return hhh + ":" + addZero(mmm);
    
} // end minute2DeltaTime

function dayMinute2HumanTime ( minute ) {
    
    if ( minute == -1 ) {
        var spc = "\u00A0\u00A0\u00A0\u00A0\u00A0";
        if ( ! use24HourClock ) {
            spc = spc + "\u00A0\u00A0";
        }
        return spc;
    }
    
    var m = Math.floor( minute );
    var h = Math.floor( m / 60.0 );
    m = m - h * 60.0;
    if ( h >= 24 ) {
        h -= 24;
    }
    var ht = addZero( h ) + ':' + addZero( m );
    if ( ! use24HourClock ) {
        ht = clock24To12( ht );
    }
    return ht;
    
}

function rgbToHex( r, g, b ) {
    
    return "#" + componentToHex(r) + componentToHex(g) + componentToHex(b);
    
} // end rgbToHex

function newTitle ( msg, minute, submsg ) {
    
    title.textContent = msg;
    var now = getNowMinute();
    var delta =  Math.round( minute - now );
    var sign = delta >= 0 ? '+' : '-';
    titleTime.textContent = sign + minute2DeltaTime( delta );
    subtitle.textContent = submsg;
    
} //end newTitle

function setMarkersVisibility( vis ) {
    
    var dawnCircle = document.getElementById( "dawnCircle" );
    var duskCircle = document.getElementById( "duskCircle" );
    var noonCircle = document.getElementById( "noonCircle" );
    var solarNoonCircle = document.getElementById( "solarNoonCircle" );
    var sunriseCircle = document.getElementById( "sunriseCircle" );
    var sunsetCircle = document.getElementById( "sunsetCircle" );
    var midnightCircle = document.getElementById( "midnightCircle" );
    var moonriseCircle = document.getElementById( "moonriseCircle" );
    var moonsetCircle = document.getElementById( "moonsetCircle" );
    var nowCircle = document.getElementById( "nowCircle" );
    dawnCircle.setAttribute( "visibility", vis );
    duskCircle.setAttribute( "visibility", vis );
    noonCircle.setAttribute( "visibility", vis );
    solarNoonCircle.setAttribute( "visibility", vis );
    sunriseCircle.setAttribute( "visibility", vis );
    sunsetCircle.setAttribute( "visibility", vis );
    midnightCircle.setAttribute( "visibility", vis );
    solarMidnightCircle.setAttribute( "visibility", vis );
    solarMidnightCircleLeft.setAttribute( "visibility", vis );
    solarMidnightCircleRight.setAttribute( "visibility", vis );
    titleTime.setAttribute( 'visibility', vis );
    moonriseCircle.setAttribute( "visibility", ( moonrise == -1 ) ? 'hidden' : vis );
    moonsetCircle.setAttribute( "visibility", ( moonset == -1 ) ? 'hidden' : vis );
    nowCircle.setAttribute( 'visibility', vis );
    
} // end setMarkersVisibility

function setDetailedInfoVisibility( vis )  {

    document.getElementById( "solarDate" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo1" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo2" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo3" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo4" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo5" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo5.1" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo6" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo7" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo8" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo9" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo9.1" ).setAttribute( 'visibility', vis );
    document.getElementById( "detailedInfo9.2" ).setAttribute( 'visibility', vis );
    document.getElementById( "youdrivetriangle" ).setAttribute( 'visibility', vis == 'visible'? 'hidden' : 'visible' );

} // setDetailedInfoVisibility

function commify(x) {
    var parts = x.toString().split(".");
    parts[0] = parts[0].replace(/\B(?=(\d{3})+(?!\d))/g, ",");
    return parts.join(".");
}

function moonElement ( id ) {
    
    var hhh = Math.floor (currentMinute / 60.0); // hour of day
    var mmm = currentMinute - (hhh*60);          // minute of day
    var now = getNowDateObject();
    now = new Date( now.getFullYear(), now.getMonth(), now.getDate(), hhh, mmm, 0, 0 );
    var trend, distance, v;

    var moonPos = SunCalc.getMoonPosition( now, latitude, longitude);
    var altitude = moonPos.altitude;
    var altitudeDegrees = altitude * ( 180 / Math.PI );
    var azimuth = moonPos.azimuth; // measured from south, moving west !!!!
    azimuth -= Math.PI;		// now normalize so we start from the north !
    if ( azimuth < 0 ) {
        azimuth += 2 * Math.PI;
    }
    var azimuthDegrees = azimuth * ( 180 / Math.PI );
    var parallacticAngle = moonPos.parallacticAngle;
    var parallacticAngleDegrees = parallacticAngle * ( 180 / Math.PI );
    [distance, trend] = Luna.earthDistance( now );
    var distanceMiles = distance * 0.62137119;
    var dString = "Distance: " + commify( parseFloat( distance ).toFixed( 0 ) ) + " km" + trend + "\n";
    var altitudeString = "Altitude: " + parseFloat( altitudeDegrees ).toFixed( 0 ) + "\xB0\n";
    var azimuthString = "Azimuth: " + parseFloat( azimuthDegrees ).toFixed( 0 ) + "\xB0\n";
    [v, trend] = Luna.earthOrbitalSpeed( distance ); // km/dy
    v /= 24;
    v = v.toFixed( 0 );
    var vString = "Velocity: " + commify(v) + " km/h"  + trend + " \n";
    var pString = "Phase: " + Luna.phase( now ) + "\n";
    var xtra = "Mass: " + Luna.mass + " kg\n";
    xtra += "Radius: " + commify( parseFloat( Luna.radius ).toFixed( 1 ) ) + " km\n";
    alert( "Luna Information\n\n" + altitudeString +  azimuthString + dString + vString + pString + xtra );
        
} // end moonElement

function sunElement ( id ) {
    
    var hhh = Math.floor (currentMinute / 60.0); // hour of day
    var mmm = currentMinute - (hhh*60);          // minute of day
    var now = getNowDateObject();
    now = new Date( now.getFullYear(), now.getMonth(), now.getDate(), hhh, mmm, 0, 0 );
    var trend, distance, v;

    var sunPos = SunCalc.getPosition( now, latitude, longitude);
    [distance, trend] = Sol.earthDistance( now );
    var altitude = sunPos.altitude;
    var altitudeDegrees = altitude * ( 180 / Math.PI );
    var azimuth = sunPos.azimuth; // measured from south, moving west !!!!
    azimuth -= Math.PI;		// now normalize so we start from the north !
    if ( azimuth < 0 ) {
        azimuth += 2 * Math.PI;
    }
    var azimuthDegrees = azimuth * ( 180 / Math.PI );
    var altitudeString = "Altitude                 :  "+  parseFloat( altitude ).toFixed( 2 ) + " rad, " + parseFloat( altitudeDegrees ).toFixed( 1 ) + " deg above the horizon\n";
    var azimuthString = "Azimuth                 :  "+  parseFloat( azimuth ).toFixed( 2 ) + " rad, " + parseFloat( azimuthDegrees ).toFixed( 1 ) + " deg clockwise from N\n";
    altitudeString = "Altitude: " + parseFloat( altitudeDegrees ).toFixed( 0 ) + "\xB0\n";
    azimuthString = "Azimuth: " + parseFloat( azimuthDegrees ).toFixed( 0 ) + "\xB0\n";
    var dString = "Distance: " + commify( parseFloat( distance ).toFixed( 0 ) ) + " km" + trend + "\n";
    [v, trend] = Sol.earthOrbitalSpeed( distance ); // km/dy
        
    var currentSolarSpeed = parseFloat( v ).toFixed( 0 );

    v /= 24;
    v = v.toFixed( 0 );
    var vString = "Velocity: " + commify(v) + " km/h"  + trend + "\n";
    var xtra = "Mass: " + Sol.mass + " kg\n"
    xtra += "Radius: " + commify( parseFloat( Sol.radius ).toFixed( 0 ) ) + " km\n";
    xtra += "Spectral Class: " + Sol.spectralClass + "\n";
    xtra += "Visual Magnitude: " + Sol.visualMagnitude + " \n";
    xtra += "Effective Temperature: " + commify( parseFloat( Sol.temperature ).toFixed( 0 ) ) + " K\n";
    xtra += "Luminosity: " + Sol.luminosity +" W\n";
    //xtra += "Metallicity: " + Sol.metallicity + "\n";
    alert( "Sol Information\n\n" +  altitudeString + azimuthString + dString + vString + xtra );
    
} // end sunElement

function triangleElement( id ) {

    var daylight = sunset - sunrise;
    var h = Math.floor( daylight / 60.0 );
    var m = daylight - h * 60.0;
    
    alert( "Daylight: " + h + " hours, " + m + " minutes"  );
    
} // end triangleElement

function lunarEclipseElement( id ) {
    
    lunarEclipses = [];
    setLunartimeperiod('LE2001');
    
    var numLunarEclipses = lunarEclipses.length;
    var msg = '<p>Between <b>Today</b> and the end of next year ';
    if ( numLunarEclipses == 0 ) {
        msg = msg + 'no lunar eclipses';
    } else {
        msg = msg + numLunarEclipses + ' lunar ' + (numLunarEclipses == 1 ? 'eclipse' : 'eclipses' );
    }
    msg = msg + ' will be visible from your current location.';
    if ( numLunarEclipses > 0 ) {
        msg = msg + "<p><ul>";
        for( var e in lunarEclipses ) {
            var eventDetails = lunarEclipses[e]; // array of 18 : 2018-Jan-31,T, 2.294, 1.315,10:51,+29,11:48,+19,12:52,+07,13:30,+01,italic-14:08,italic--06,italic-15:11,italic--15,italic-16:08,italic--23
            var event = eType[ eventDetails[1] ] + ' Eclipse ' + lunarHuman2solarHumanDate( eventDetails[0] );
            msg = msg + '<li>' + event;
        }
    }
    
    if ( numLunarEclipses > 0 ) {
        // Must recalulate everything now that we have a real date so that the timezone is accurate.
        var month2Ord = {
            'Jan' : '01',
            'Feb' : '02',
            'Mar' : '03',
            'Apr' : '04',
            'May' : '05',
            'Jun' : '06',
            'Jul' : '07',
            'Aug' : '08',
            'Sep' : '09',
            'Oct' : '10',
            'Nov' : '11',
            'Dec' : '12',
        };
        var le = lunarEclipses[0];
        debugDate = le[0].substring(0,4) + '-' + month2Ord[ le[0].substring(5,8) ] + '-' + le[0].substring(9,11);
        lunarEclipses = [];
        setLunartimeperiod('LE2001');
    }
        
    if ( screenSaver == 0 ) {
        msg = msg + "</ul><p>" + showLunarEclipseData( 1 );
        var help = document.getElementById( 'modal-content-here' );
        help.style['color'] = 'black';
        myAlert( 'LunarEclipseIcon-1200x1200.png', msg );
    }

} // end lunarEclipseElement

function lunarHuman2solarHumanDate (lunarStyle) {
    
    // 2018-Jan-31
    // 2
    // 31 January 2018
    
    var y = lunarStyle.substring(0, 4);
    var m = lunarStyle.substring(5, 8);
    var d = lunarStyle.substring(9, 11);
    m = fullMonth1[ m ];
    if ( d.substring(0, 1) == '0' ) {
        d = d.substring(1, 2);
    }
    
    return d + ' ' + m + ' ' + y;

} // end lunarHuman2solarHumanDate

function eclipseElement( id ) {

    solarEclipses = [];
    setSolartimeperiod('SE2001');
    
    var numSolarEclipses = solarEclipses.length;
    var msg = '<p>Between <b>Today</b> and the end of next year ';
    if ( numSolarEclipses == 0 ) {
        msg = msg + 'no solar eclipses';
    } else {
        msg = msg + numSolarEclipses + ' solar ' + (numSolarEclipses == 1 ? 'eclipse' : 'eclipses' );
    }
    msg = msg + ' will be visible from your current location.';
    if ( numSolarEclipses > 0 ) {
        msg = msg + "<p><ul>";
        for( var e in solarEclipses ) {
            var eventDetails = solarEclipses[e]; // array of 16 : 2017-Aug-21,P,11:38:05,52,133,-,13:03:49,60,168,-,14:29:57,58,209,0.985,0.987,-
            var event = eTypeSolar[ eventDetails[1] ] + ' Eclipse ' + lunarHuman2solarHumanDate( eventDetails[0] );
            msg = msg + '<li>' + event;
        }
    }
    
    if ( numSolarEclipses > 0 ) {
        // Must recalulate everything now that we have a real date so that the timezone is accurate.
        var month2Ord = {
            'Jan' : '01',
            'Feb' : '02',
            'Mar' : '03',
            'Apr' : '04',
            'May' : '05',
            'Jun' : '06',
            'Jul' : '07',
            'Aug' : '08',
            'Sep' : '09',
            'Oct' : '10',
            'Nov' : '11',
            'Dec' : '12',
        };
        var le = solarEclipses[0];
        debugDate = le[0].substring(0,4) + '-' + month2Ord[ le[0].substring(5,8) ] + '-' + le[0].substring(9,11);
        solarEclipses = [];
        setSolartimeperiod('SE2001');
    }

    if ( screenSaver == 0 ) {
        msg = msg + "</ul><p>" + showSolarEclipseData( 1 );
        var help = document.getElementById( 'modal-content-here' );
        help.style['color'] = 'black';
        myAlert( 'SolarEclipseIcon-1200x1200.png', msg );
    }
    
    return;
    
} // end eclipseElement

// New day animation.

function hideAllLunarElements () {
    
    setLunaVis( 'hidden' );
    lunaSinePath.setAttribute( 'opacity', 0.0 );
    lunaPath.setAttribute( 'opacity', 0.0 );
    lunabackPath.setAttribute( 'opacity', 0.0 );

} // end hideAllLunarElements

//
// Tween functions.
//

function playMissedMinutes ( minute ) {
    
    // Tween missed minutes: forward and reverse ease-in and ease-out the simulation.
    // I learned all this from http://upshots.org/actionscript/jsas-understanding-easing.
        
        //console.log("Start Tween minute=" + minute + ", LNM=" + lastKnownMinute + ", tweenTime=" + tweenTime);
        //if ( minute < 0 || minute >= mpd ) minute = 0;
        //if ( lastKnownMinute < 0 || lastKnownMinute >= mpd ) lastKnownMinute = 0;

    if ( lastKnownMinute < ( minute - minutesInterval ) ) { // if before now
        if ( tweenTime == -1 ) {
            tweenTime = 0;
            tweenStart = lastKnownMinute;
            tweenEnd = minute - lastKnownMinute;
            tweenTimeEnd = ( tweenEnd < mpd / 2 ? 1 : 2 );
            //console.log( "start forward tween, LNM="+lastKnownMinute+", minute="+minute+", tweenStart="+tweenStart+", tweenEnd="+tweenEnd+", tweenTineEnd="+tweenTimeEnd );
        }
        minute = tween( tweenTime, tweenStart, tweenEnd, tweenTimeEnd );
        minute = Math.round( minute );
        tweenTime += tweenDT / ( 1000 * minutesInterval );
        lastKnownMinute = minute + minutesInterval;
    } else 	if ( lastKnownMinute > ( minute + minutesInterval ) ) { // if after now
        if ( tweenTime == -1 ) {
            tweenTime = 0;
            tweenStart =  minute;
            //tweenEnd = lastKnownMinute - minute;
            tweenEnd = ( ( lastKnownMinute - minute ) % mpd ) + mpd % mpd;
            tweenTimeEnd = ( tweenEnd < mpd / 2 ? 1 : 2 );
            //console.log( "start reverse tween, LNM="+lastKnownMinute+", minute="+minute+", tweenStart="+tweenStart+", tweenEnd="+tweenEnd+", tweenTineEnd="+tweenTimeEnd );
        }
        minute = tween( tweenTime, tweenStart, tweenEnd, tweenTimeEnd );
        minute = ( tweenStart + tweenEnd ) - ( minute - tweenStart );
        minute = Math.round( minute );
        tweenTime += tweenDT / ( 1000 * minutesInterval );
        lastKnownMinute = minute - minutesInterval;
    } else {
        var d = new Date();
        console.log("Tween bug at "+d+", LNM="+lastKnownMinute+", minute="+minute);
        //lastKnownMinute = minute - 1;
        lastKnownMinute = getNowMinute();
        tweenTime = -1;
    }
    
    return Math.round( minute );
    
} // end playMissedMinutes

function tween ( t, b, c, d ) {
    
    //    return c * t / d + b;         // linear
    //    return c*((t=t/d-1)*t*t+1)+b; // ease out cubic
    //
    // Ease-in and ease-out quadratic.
    
    if ( (t /= d / 2 ) < 1 ) {
        return c / 2 * t * t + b;
    }
    return -c / 2 * ( (--t) * ( t - 2 ) -1 ) + b;
    
} // end tween

//
// Lunar eclipse funtions.
//

function closeMyModal() {
    
    var modalWindow = document.getElementById('myModalWindow');
    modalWindow.style.display = "none";
    
}

function myAlert( img, msg ) {
    
    if ( myAlertActive ) {
        alert("myAlert() is already active, this should not happen");
        return;
    }
    myAlertActive = 1;
    
    var iconWidth = 64;
    var iconHeight = 64;
    if ( 1 ) { // if iOS
        iconWidth *= 2;
        iconHeight *= 2;
    }
    var leftPadding = (iOSInfoMsg == '' ? 2 : 0);
    var newMsg = '<button id="myModalOK" onclick="closeMyModal();" style="background:red; text-align:center; width:36px; height:36px; border-radius:100%; border:1px red; display: inline-flex;flex-direction: column; justify-content:center; align-items:center; font-size:28px; color: black; padding: 0px 0px 0px ' + leftPadding + 'px; text-decoration:none; " >X</button><p>';
    if ( img != '' ) {
        newMsg = newMsg + '<img src="imgs/' + img + '" width=' + iconWidth + 'px height=' + iconHeight + 'px> <br>';
    }
    var modalWindow = document.getElementById('myModalWindow');
    var modalContent = document.getElementById('modal-content-here');
    modalWindow.style.display = "block";
    modalContent.innerHTML = newMsg + msg;

} // end myAlert

function checkForLunarEclipse ( ) {
    
    // If an entry in lunarEclipses[] matches today, then display its details.
    
    lunarEclipses = [];
    setLunartimeperiod( "LE2001" );
    var msg = showLunarEclipseData( 0 );
    if ( msg != '' && msg != '</table>' && screenSaver == 0 ) {
        var help = document.getElementById( 'modal-content-here' );
        help.style['color'] = 'black';
        myAlert( 'LunarEclipseIcon-1200x1200.png', msg );
    }

} // end checkForLunarEclipse

function getLunarAzimuthDegrees ( hhmm, penumbralEclipseBegins ) {

    var toks = debugDate.split( '-' );
    hhmm = hhmm.replace( 'italic-', '' );
    var dt = new Date( toks[0], toks[1]-1, toks[2], hhmm.substr(0,2), hhmm.substr(3,2), 0, 0 );
    var penumbralDate = new Date( toks[0], toks[1]-1, toks[2], penumbralEclipseBegins.substr(0,2), penumbralEclipseBegins.substr(3,2), 0, 0 );
    if ( dt.getTime() < penumbralDate.getTime() ) { // if current time is < penumbral begins, then we are in the next day
        var nextDay = new Date(dt);
        nextDay.setDate(dt.getDate()+1);
        dt = nextDay;
    }
    var moonPos = SunCalc.getMoonPosition( dt, latitude, longitude);
    var lunarAzimuth = moonPos.azimuth; // measured from south, moving west !!!!
    lunarAzimuth -= Math.PI;        // now normalize so we start from the north !
    if ( lunarAzimuth < 0 ) {
        lunarAzimuth += 2 * Math.PI;
    }
    lunarAzimuthDegrees = lunarAzimuth * ( 180 / Math.PI );
    return lunarAzimuthDegrees;

} // end getLunarAzimuthDegrees

function joinAltAzi ( hhmm, altitude, penumbralEclipseBegins ) {
    
    var alt = tidyDegrees( altitude );
    var azi = '-';
    if ( hhmm != '-') {
        azi= parseFloat( getLunarAzimuthDegrees( hhmm, penumbralEclipseBegins ) ).toFixed(  ) + "\xB0";
    }
    return '(' + alt + ',' + azi + ')';
    
} // end joinAltAzi

function computeLunarEclipseDuration ( s, e ) {
    
    if ( s =='-' || e == '-' ) {
        return '';
    }
    
    var end = e.replace( 'italic-', '' );
    var start = s.replace( 'italic-', '' );
    end = end.split(':');
    start = start.split(':');
    var endSecs = end[0]*3600 + end[1]*60 + (+end[2] || 0);
    var startSecs = start[0]*3600 + start[1]*60 + (+start[2] || 0);
    var secs = endSecs - startSecs;
    if ( secs < 0 ) {
        secs += 86400;
    }
    return (secs/3600 |0) + 'h ' + ((secs%3600) / 60 |0) + 'm '; //  + (secs%60) + 's';

}

function showLunarEclipseData( canResetDebugDate ) {
    
    var msg = '';
    var numLunarEclipses = lunarEclipses.length
    if ( numLunarEclipses <= 0 ) {
        return msg;
    }
    var now = getNowDateObject();
    var ly = now.getFullYear();
    var lm = now.getMonth();
    lm = month[ lm ];
    var ld = now.getDate();
    ld = addZero( ld );
    var ldate = ly + '-' + lm + '-' + ld;
    for ( i = 0; i < numLunarEclipses; i++ ) {
        var le = lunarEclipses[ i ];
        if ( le[0] == ldate || canResetDebugDate ) {
            if ( canResetDebugDate ) {
                var month2Ord = {
                    'Jan' : '01',
                    'Feb' : '02',
                    'Mar' : '03',
                    'Apr' : '04',
                    'May' : '05',
                    'Jun' : '06',
                    'Jul' : '07',
                    'Aug' : '08',
                    'Sep' : '09',
                    'Oct' : '10',
                    'Nov' : '11',
                    'Dec' : '12',
                };
                if ( canResetDebugDate ) {
                    debugDate = le[0].substring(0,4) + '-' + month2Ord[ le[0].substring(5,8) ] + '-' + le[0].substring(9,11);
                }
                lm = le[0].substring(5,8);
                ldate = le[0].substring(0,4) + '-' + lm + '-' + le[0].substring(9,11);
                if ( screenSaver == 0 ) {
                    msg = msg + "<p>The <b>You Drive</b> simulator date has been temporarily set to the eclipse date '" + debugDate + "'.";
                    var nowMin = getNowMinute();
                    updateRiseAndSetTimes( nowMin );
                    animateSol ( nowMin );
                    animateLuna ( nowMin  );
                }
            }
            msg = msg + '<center><h1>Lunar Eclipse of ' + lunarHuman2solarHumanDate( ldate ) + "</h1>";
            msg = msg + '<h1>Moon in ' + eType[ le[1] ] + " Eclipse at this Location</h1></center>\n";
            msg = msg + "<table border=0>\n";
            var duration = computeLunarEclipseDuration( le[ 8], le[12] );
            if ( duration != '' ) {
                msg += formatTD ( 'Total duration',  duration );
            }
            duration = computeLunarEclipseDuration( le[ 6], le[14] );
            if ( duration != '' ) {
                msg += formatTD ( 'Partial duration',  duration );
            }
            duration = computeLunarEclipseDuration( le[ 4], le[16] );
            if ( duration != '' ) {
                msg += formatTD ( 'Penumbral duration',  duration );
            }
            msg += formatTD( '-', '-' );
            msg = msg + formatTD ( 'Penumbral magnitude',             le[ 2] );
            msg = msg + formatTD ( 'Umbral magnitude',                le[ 3] );
            msg += formatTD( '-', '-' );
            if ( le[ 4] != '-' ) {
                msg = msg + formatTD ( 'Penumbral eclipse begins',        le[ 4] );
                msg = msg + formatTD ( '(Alti,Azim)',         joinAltAzi( le[ 4], le[ 5], le[ 4] ) );
            }
            if ( le[ 6] != '-' ) {
                msg = msg + formatTD ( 'Partial eclipse begins',          le[ 6] );
                msg = msg + formatTD ( '(Alti,Azim)',         joinAltAzi( le[ 6], le[ 7], le[ 4] ) );
            }
            if ( le[ 8] != '-' ) {
                msg = msg + formatTD ( 'Total eclipse begins',            le[ 8] );
                msg = msg + formatTD ( '(Alti,Azim)',         joinAltAzi( le[ 8], le[ 9], le[ 4] ) );
            }
            if ( le[10] != '-' ) {
                msg = msg + formatTD ( 'Middle eclipse',                  le[10] );
                msg = msg + formatTD ( '(Alti,Azim)',         joinAltAzi( le[10], le[11], le[ 4] ) );
            }
            if ( le[12] != '-' ) {
                msg = msg + formatTD ( 'Total eclipse ends',              le[12] );
                msg = msg + formatTD ( '(Alti,Azim)',         joinAltAzi( le[12], le[13], le[ 4] ) );
            }
            if ( le[14] != '-' ) {
                msg = msg + formatTD ( 'Partial eclipse ends',            le[14] );
                msg = msg + formatTD ( '(Alti,Azim)',         joinAltAzi( le[14], le[15], le[ 4] ) );
            }
            if ( le[16] != '-' ) {
                msg = msg + formatTD ( 'Penumbral eclipse ends',          le[16] );
                msg = msg + formatTD ( '(Alti,Azim)',         joinAltAzi( le[16], le[17], le[ 4] ) );
            }
            break;
        }
    }
    msg = msg + '</table>';
    return msg;
    
} // end showLunarEclipseData

function tidyDegrees ( le ) { // tidy degrees
    
    var d = le;
    if ( d == '-' ) {
        return d;
    }
    d = d.replace( '+', '');
    if ( d == '0(r)' ) {
        return '0' + "\xb0" + '(r)';
    }
    if ( d == '0(s)' ) {
        return '0' + "\xb0" + '(s)';
    }
    if ( d.length > 1 && d.substr(0, 1) == '0' ) {
        d = d.substr( 1, d.length - 1 );
    }
    if ( d.length > 2 && d.substr(0, 2) == '-0' ) {
        d = '-' + d.substr( 2, d.length - 2 );
    }
    // italic-05
    if ( d.length > 1+7 && d.substr(0, 1+7) == 'italic-0' ) {
        d = 'italic-' + d.substr( 1+7, d.length - 1+7 );
    }
    // italic--05
    if ( d.length > 1+7+1 && d.substr(0, 1+7+1) == 'italic--0' ) {
        d = 'italic--' + d.substr( 1+7+1, d.length - 1+7+1 );
    }

    d += "\xB0";
    return d;
    
}//end tidyDegrees

function formatTD ( desc, val ) {
    
    if ( desc == '-' && val =='-' ) {
        return "<tr> <td align=right> <font size=6>&nbsp</font></td>  <td align=right> <font size=6></td> </font> </tr>\n";
    }
    
    var newVal;
    var italicize = 0;
    if ( val.search( 'italic-') == -1 ) {
        newVal = val;
    } else {
        newVal = val.replace( 'italic-', '' );
        italicize = 1;
    }
    
    // Covert to AM / PM if that preference is set:
    //      Lunar time format is hh:mm
    //      Solar time format is hh:mm:ss and hh:mm:ss.fraction
    // There may be an (r) or (s)
    
    var rs = '';
    if ( ! use24HourClock ) {
        if ( newVal.substr( newVal.length - 3, 3 ) == '(r)' ) {
            rs = '(r)';
            newVal = newVal.replace( '(r)', '' );
            if ( newVal.length == 5 ) {
                newVal += ':00';
            }
        }
        if ( newVal.substr( newVal.length - 3, 3 ) == '(s)' ) {
            rs = '(s)';
            newVal = newVal.replace( '(s)', '' );
            if ( newVal.length == 5 ) {
                newVal += ':00';
            }
        }
        if ( newVal.length == 5 && newVal.substr(2,1) == ':' ) {
            var h = newVal.substr(0,2);
            var pm = ' am';
            if ( h > 12 ) {
                h -= 12;
                pm = ' pm';
            }
            newVal = h + ':' + newVal.substr(3,2) + pm;
        } else if ( newVal.length == 10 && newVal.substr(2,1) == ':' && newVal.substr(5,1) == ':' && newVal.substr(8,1) == '.' ) { // 11:38:07.2
            var h = newVal.substr(0,2);
            var pm = ' am';
            if ( h > 12 ) {
                h -= 12;
                pm = ' pm';
            }
            newVal = h + ':' + newVal.substr(3,5) + pm;
        } else if ( newVal.length == 8 && newVal.substr(2,1) == ':' && newVal.substr(5,1) == ':' ) { // 13:03:49
            var h = newVal.substr(0,2);
            var pm = ' am';
            if ( h > 12 ) {
                h -= 12;
                pm = ' pm';
            }
            newVal = h + ':' + newVal.substr(3,5) + pm;
        }

    }
    
    newVal += rs;

    if ( italicize ) {
        newVal = '<font color=#787878><i>' + newVal + '</i></font>';
    }
    
    return '<tr> <td align=right> <font size=6>' + desc + ':&nbsp</font></td>  <td align=right> <font size=6>' + newVal + "</td> </font> </tr>\n";
    
} // end formatTD

//
// Solar eclipse functions.
//

function checkForSolarEclipse ( ) {
    
    // If an entry in solarEclipses[] matches today, then display its details.
    
    solarEclipses = [];
    setSolartimeperiod( "SE2001" );
    var msg = showSolarEclipseData( 0 );
    if ( msg != '' && msg != '</table>' && screenSaver == 0) {
        var help = document.getElementById( 'modal-content-here' );
        help.style['color'] = 'black';
        myAlert( 'SolarEclipseIcon-1200x1200.png', msg );
    }

    return;
    
} // end checkForSolarEclipse

function joinSolarAltAzi ( altitude, azimuth ) {
    
    var alt = tidyDegrees( altitude );
    var azi = tidyDegrees( azimuth );
    return '(' + alt + ',' + azi + ')';
    
} // end joinAltAzi

function showSolarEclipseData( canResetDebugDate ) {
    
    var msg = '';
    var numSolarEclipses = solarEclipses.length
    if ( numSolarEclipses <= 0 ) {
        return msg;
    }
    var now = getNowDateObject();
    var ly = now.getFullYear();
    var lm = now.getMonth();
    lm = month[ lm ];
    var ld = now.getDate();
    ld = addZero( ld );
    var ldate = ly + '-' + lm + '-' + ld;
    for ( i = 0; i < numSolarEclipses; i++ ) {
        var le = solarEclipses[ i ];
        if ( le[0] == ldate || canResetDebugDate ) {
            if ( canResetDebugDate ) {
                var month2Ord = {
                    'Jan' : '01',
                    'Feb' : '02',
                    'Mar' : '03',
                    'Apr' : '04',
                    'May' : '05',
                    'Jun' : '06',
                    'Jul' : '07',
                    'Aug' : '08',
                    'Sep' : '09',
                    'Oct' : '10',
                    'Nov' : '11',
                    'Dec' : '12',
                };
                if ( canResetDebugDate ) {
                    debugDate = le[0].substring(0,4) + '-' + month2Ord[ le[0].substring(5,8) ] + '-' + le[0].substring(9,11);
                }
                lm = le[0].substring(5,8);
                ldate = le[0].substring(0,4) + '-' + lm + '-' + le[0].substring(9,11);
                if ( screenSaver == 0 ) {
                    msg = msg + "<p>The <b>You Drive</b> simulator date has been temporarily set to the eclipse date '" + debugDate + "'.";
                    var nowMin = getNowMinute();
                    updateRiseAndSetTimes( nowMin );
                    animateSol ( nowMin );
                    animateLuna ( nowMin  );
                }
            }
            msg = msg + '<center><h1>Solar Eclipse of ' + lunarHuman2solarHumanDate( ldate ) + "</h1>";
            msg = msg + '<h1>Sun in ' + eTypeSolar[ le[1] ] + " Eclipse at this Location</h1></center>\n";
            msg = msg + "<table border=0>\n";
            var duration = "";
            var end = le[10].replace( '(s)', '' );
            end = end.split(':');
            var start = le[2].replace( '(r)', '' );
            start = start.split(':');
            var endSecs = end[0]*3600 + end[1]*60 + (+end[2] || 0);
            var startSecs = start[0]*3600 + start[1]*60 + (+start[2] || 0);
            var secs = endSecs - startSecs;
            duration = (secs/3600 |0) + 'h ' + ((secs%3600) / 60 |0) + 'm ' + (secs%60) + 's';
            msg = msg + formatTD ( 'Duration',                              duration );
            msg = msg + formatTD ( 'Magnitude',                             le[13] );
            var rs = '';
            var obscuration = le[14];
            if ( obscuration.endsWith( '(s)' ) || obscuration.endsWith( '(r)' ) ) {
                rs = obscuration.substr( obscuration.length-3, 3 );
                obscuration = obscuration.substr( 0, obscuration.length-3 );
            }
            obscuration = ( obscuration * 100.0 ) + '%' + rs;
            msg = msg + formatTD ( 'Obscuration',                           obscuration );
            msg += formatTD( '-', '-' );
            msg = msg + formatTD( "Solar angular diameter", parseFloat( Sol.angularDiameterFromEarth( now ) ).toFixed( 3 ) + "\xB0" );
            msg = msg + formatTD( "Lunar angular diameter", parseFloat( Luna.angularDiameterFromEarth( now ) ).toFixed( 3 ) + "\xB0" );
            msg += formatTD( '-', '-' );
            msg = msg + formatTD ( 'Partial eclipse begins',                le[ 2] );
            msg = msg + formatTD ( '(Alti,Azim)',          joinSolarAltAzi( le[ 3], le[4] ) );
            if ( le[ 5] != '-' ) msg = msg + formatTD ( 'A or T eclipse begins',                 le[ 5] );
            msg = msg + formatTD ( 'Maximum eclipse',                       le[ 6] );
            msg = msg + formatTD ( '(Alti,Azim)',          joinSolarAltAzi( le[ 7], le[ 8] ) );
            if ( le[ 9] != '-' ) msg = msg + formatTD ( 'A or T eclipse ends',                   le[ 9] );
            msg = msg + formatTD ( 'Partial eclipse ends',                  le[10] );
            msg = msg + formatTD ( '(Alti,Azim)',          joinSolarAltAzi( le[11], le[12] ) );
            if ( le[15] != '-' ) msg = msg + formatTD ( 'A or T eclipse duration',               le[15] );
            msg = msg + "</table>\n";
            msg = msg + "<p>Abbreviations: A = Annular, T = Total, (r) = in progress at sun<u>r</u>ise, (s) = in progress at sun<u>s</u>et.";
            break;
        }
    }
    msg = msg + '</table>';
    return msg;
    
} // showSolarEclipseData
